import re

import numpy as np

from ..geometry import triangulate_quads
from ..util import array_to_string, comment_strip, decode_text


def load_off(file_obj, **kwargs) -> dict:
    """
    Load an OFF file into the kwargs for a Trimesh constructor.

    Parameters
    ----------
    file_obj : file object
      Contains an OFF file

    Returns
    ----------
    loaded : dict
      kwargs for Trimesh constructor
    """
    text = file_obj.read()
    # will magically survive weird encoding sometimes
    # comment strip will handle all cases of commenting
    text = comment_strip(decode_text(text)).strip()

    # split the first key
    _, header, raw = re.split("(COFF|OFF)", text, maxsplit=1)
    if header.upper() not in ["OFF", "COFF"]:
        raise NameError(f"Not an OFF file! Header was: `{header}`")

    # split into lines and remove whitespace
    splits = [i.strip() for i in str.splitlines(str(raw))]
    # remove empty lines
    splits = [i for i in splits if len(i) > 0]

    # the first non-comment line should be the counts
    header = np.array(splits[0].split(), dtype=np.int64)
    vertex_count, face_count = header[:2]

    vertices = np.array(
        [i.split()[:3] for i in splits[1 : vertex_count + 1]], dtype=np.float64
    )

    # will fail if incorrect number of vertices loaded
    vertices = vertices.reshape((vertex_count, 3))

    # get lines with face data
    faces = [i.split() for i in splits[vertex_count + 1 : vertex_count + face_count + 1]]
    # the first value is count
    faces = [line[1 : int(line[0]) + 1] for line in faces]

    faces = triangulate_quads(faces)
    # save data as kwargs for a trimesh.Trimesh
    kwargs = {"vertices": vertices, "faces": faces}

    return kwargs


def export_off(mesh, digits=10) -> str:
    """
    Export a mesh as an OFF file, a simple text format

    Parameters
    -----------
    mesh : trimesh.Trimesh
      Geometry to export
    digits : int
      Number of digits to include on floats

    Returns
    -----------
    export : str
      OFF format output
    """
    # make sure specified digits is an int
    digits = int(digits)
    # prepend a 3 (face count) to each face
    faces_stacked = np.column_stack((np.ones(len(mesh.faces)) * 3, mesh.faces)).astype(
        np.int64
    )
    # the header is vertex count, face count, another number
    export = "\n".join(
        [
            "OFF",
            str(len(mesh.vertices)) + " " + str(len(mesh.faces)) + " 0",
            array_to_string(mesh.vertices, col_delim=" ", row_delim="\n", digits=digits),
            array_to_string(faces_stacked, col_delim=" ", row_delim="\n"),
            "",
        ]
    )

    return export


_off_loaders = {"off": load_off}
_off_exporters = {"off": export_off}



import os
import platform

from .. import util
from ..constants import log
from ..resources import get_string
from ..typed import BooleanOperationType, Iterable
from .generic import MeshScript

if platform.system() == "Windows":
    # try to find Blender install on Windows
    # split existing path by delimiter
    _search_path = [i for i in os.environ.get("PATH", "").split(";") if len(i) > 0]
    for pf in [r"C:\Program Files", r"C:\Program Files (x86)"]:
        pf = os.path.join(pf, "Blender Foundation")
        if os.path.exists(pf):
            for p in os.listdir(pf):
                if "Blender" in p:
                    _search_path.append(os.path.join(pf, p))
    _search_path = ";".join(set(_search_path))
    log.debug("searching for blender in: %s", _search_path)
elif platform.system() == "Darwin":
    # try to find Blender on Mac OSX
    _search_path = [i for i in os.environ.get("PATH", "").split(":") if len(i) > 0]
    _search_path.extend(
        [
            "/Applications/blender.app/Contents/MacOS",
            "/Applications/Blender.app/Contents/MacOS",
            "/Applications/Blender/blender.app/Contents/MacOS",
        ]
    )
    _search_path = ":".join(set(_search_path))
    log.debug("searching for blender in: %s", _search_path)
else:
    _search_path = os.environ.get("PATH", "")

_blender_executable = util.which("blender", path=_search_path)
exists = _blender_executable is not None

# a map that translates:
# `trimesh.BooleanOperationType` -> blender value
_blender_bool = {
    "union": "UNION",
    "difference": "DIFFERENCE",
    "intersection": "INTERSECT",
}


def boolean(
    meshes: Iterable,
    operation: BooleanOperationType = "difference",
    use_exact: bool = True,
    use_self: bool = False,
    debug: bool = False,
    check_volume: bool = True,
):
    """
    Run a boolean operation with multiple meshes using Blender.

    Parameters
    -----------
    meshes
      List of mesh objects to be operated on
    operation
      Type of boolean operation ("difference", "union", "intersect").
    use_exact
      Use the "exact" mode as opposed to the "fast" mode.
    use_self
      Whether to consider self-intersections.
    debug
      Provide additional output for troubleshooting.
    check_volume
      Raise an error if not all meshes are watertight
      positive volumes. Advanced users may want to ignore
      this check as it is expensive.

    Returns
    ----------
    result
      The result of the boolean operation on the provided meshes.
    """
    if not exists:
        raise ValueError("No blender available!")
    if check_volume and not all(m.is_volume for m in meshes):
        raise ValueError("Not all meshes are volumes!")

    # conversions from the trimesh `BooleanOperationType` to the blender option
    key = operation.lower().strip()
    if key not in _blender_bool:
        raise ValueError(
            f"`{operation}` is not a valid boolean: `{_blender_bool.keys()}`"
        )

    if use_exact:
        solver_options = "EXACT"
    else:
        solver_options = "FAST"

    # get the template from our resources folder
    template = get_string("templates/blender_boolean.py.tmpl")
    # use string substitutions rather than `string.Template` as we aren't going
    # to be filling in all the values here, `MeshScript` is going to be
    # the source of `$MESH_PRE`, etc.
    script = (
        template.replace("$OPERATION", _blender_bool[key])
        .replace("$SOLVER_OPTIONS", solver_options)
        .replace("$USE_SELF", f"{use_self}")
    )
    with MeshScript(meshes=meshes, script=script, debug=debug) as blend:
        result = blend.run(_blender_executable + " --background --python $SCRIPT")

    result = util.make_sequence(result)
    for m in result:
        # blender returns actively incorrect face normals
        m.face_normals = None

    return util.concatenate(result)


def unwrap(
    mesh, angle_limit: float = 66.0, island_margin: float = 0.0, debug: bool = False
):
    """
    Run an unwrap operation using blender.
    """
    if not exists:
        raise ValueError("No blender available!")

    # get the template from our resources folder
    template = get_string("templates/blender_unwrap.py.template")
    script = template.replace("$ANGLE_LIMIT", f"{angle_limit:.6f}").replace(
        "$ISLAND_MARGIN", f"{island_margin:.6f}"
    )

    with MeshScript(meshes=[mesh], script=script, exchange="obj", debug=debug) as blend:
        result = blend.run(_blender_executable + " --background --python $SCRIPT")

    for m in util.make_sequence(result):
        # blender returns actively incorrect face normals
        m.face_normals = None

    return result



try:
    from . import generic as g
except BaseException:
    import generic as g


def test_units():
    # make sure unit conversions fail for fake units
    fake_units = "blorbs"
    fake_units = "in"
    try:
        c = g.trimesh.units.unit_conversion("inches", fake_units)  # NOQA
        raise AssertionError()
    except BaseException:
        pass

    m = g.get_mesh("featuretype.STL")
    assert m.units is None

    m.units = "in"
    assert m.units == "in"

    extents_pre = m.extents
    m.convert_units("mm")
    scale = g.np.divide(m.extents, extents_pre)
    assert g.np.allclose(scale, 25.4)
    assert m.units == "mm"


def test_mm_hint():
    # get an STL that has "mm" in the name
    m = g.get_mesh("20mm-xyz-cube.stl")
    # convert it to meters
    m.convert_units("m")
    # should have picked the millimeters from the name
    assert g.np.allclose(m.extents, 0.02)


def test_conversion():
    # test conversions on a multibody STL in a scene

    # a multibody STL with a unit hint in filename
    m = g.get_mesh("counter.unitsmm.STL")

    # nothing should be set
    assert m.units is None

    # split into watertight bodies
    s = g.trimesh.scene.split_scene(m)

    # save the extents
    extents_pre = s.extents

    # should extract units from file name without
    # raising a ValueError
    c = s.convert_units("in", guess=False)
    # should have converted mm -> in, 1/25.4
    # extents should scale exactly with unit conversion
    assert g.np.allclose(extents_pre / c.extents, 25.4, atol=0.01)


def test_path():
    p = g.get_mesh("2D/tray-easy1.dxf")
    # should be inches
    assert "in" in p.units
    extents_pre = p.extents
    p.convert_units("mm")
    # should have converted in -> mm 25.4
    # extents should scale exactly with unit conversion
    assert g.np.allclose(p.extents / extents_pre, 25.4, atol=0.01)


def test_keys():
    units = g.trimesh.units.keys()
    assert isinstance(units, set)
    assert "in" in units


def test_arbitrary():
    ac = g.np.allclose
    to_inch = g.trimesh.units.to_inch

    # check whitespace
    assert ac(1.0, to_inch("in"))
    assert ac(1.0, to_inch("1.00000* in"))
    assert ac(1.0, to_inch("1.00    * in"))

    # check centimeter conversion
    assert ac(100, to_inch("m") / to_inch("0.01*m"))

    # if we are currently in centimeters and want to go to meters
    # it should be dividing it by 100
    assert ac(0.01, g.trimesh.units.unit_conversion(current="0.01*m", desired="meters"))


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    test_mm_hint()



"""
grouping.py
-------------

Functions for grouping values and rows.
"""

import numpy as np

from . import util
from .constants import log, tol
from .typed import ArrayLike, Integer, NDArray, Number, Optional, Sequence, Tuple

try:
    from scipy.spatial import cKDTree
except BaseException as E:
    # wrapping just ImportError fails in some cases
    # will raise the error when someone tries to use KDtree
    from . import exceptions

    cKDTree = exceptions.ExceptionWrapper(E)


def merge_vertices(
    mesh,
    merge_tex: Optional[bool] = None,
    merge_norm: Optional[bool] = None,
    digits_vertex: Optional[Integer] = None,
    digits_norm: Optional[Integer] = None,
    digits_uv: Optional[Integer] = None,
):
    """
    Removes duplicate vertices, grouped by position and
    optionally texture coordinate and normal.

    Parameters
    -------------
    mesh : Trimesh object
      Mesh to merge vertices on
    merge_tex : bool
      If True textured meshes with UV coordinates will
      have vertices merged regardless of UV coordinates
    merge_norm : bool
      If True, meshes with vertex normals will have
      vertices merged ignoring different normals
    digits_vertex : None or int
      Number of digits to consider for vertex position
    digits_norm : int
      Number of digits to consider for unit normals
    digits_uv : int
      Number of digits to consider for UV coordinates
    """
    # no vertices so exit early
    if len(mesh.vertices) == 0:
        return
    if merge_tex is None:
        merge_tex = False
    if merge_norm is None:
        merge_norm = False
    if digits_norm is None:
        digits_norm = 2
    if digits_uv is None:
        digits_uv = 4
    if digits_vertex is None:
        # use tol.merge if digit precision not passed
        digits_vertex = util.decimal_to_digits(tol.merge)

    # if we have a ton of unreferenced vertices it will
    # make the unique_rows call super slow so cull first
    if hasattr(mesh, "faces") and len(mesh.faces) > 0:
        referenced = np.zeros(len(mesh.vertices), dtype=bool)
        referenced[mesh.faces] = True
    else:
        # this is used for geometry without faces
        referenced = np.ones(len(mesh.vertices), dtype=bool)

    # collect vertex attributes into sequence we can stack
    stacked = [mesh.vertices * (10**digits_vertex)]

    # UV texture visuals require us to update the
    # vertices and normals differently
    if (
        not merge_tex
        and mesh.visual.defined
        and mesh.visual.kind == "texture"
        and mesh.visual.uv is not None
        and len(mesh.visual.uv) == len(mesh.vertices)
    ):
        # get an array with vertices and UV coordinates
        # converted to integers at requested precision
        stacked.append(mesh.visual.uv * (10**digits_uv))

    # check to see if we have vertex normals
    normals = mesh._cache["vertex_normals"]
    if not merge_norm and np.shape(normals) == mesh.vertices.shape:
        stacked.append(normals * (10**digits_norm))

    # stack collected vertex properties and round to integer
    stacked = np.column_stack(stacked).round().astype(np.int64)

    # check unique rows of referenced vertices
    u, i = unique_rows(stacked[referenced], keep_order=True)

    # construct an inverse using the subset
    inverse = np.zeros(len(mesh.vertices), dtype=np.int64)
    inverse[referenced] = i
    # get the vertex mask
    mask = np.nonzero(referenced)[0][u]
    # run the update including normals and UV coordinates
    mesh.update_vertices(mask=mask, inverse=inverse)


def group(values, min_len: Optional[Integer] = None, max_len: Optional[Integer] = None):
    """
    Return the indices of values that are identical

    Parameters
    ----------
    values : (n,) int
      Values to group
    min_len : int
      The shortest group allowed
      All groups will have len >= min_length
    max_len : int
      The longest group allowed
      All groups will have len <= max_length

    Returns
    ----------
    groups : sequence
      Contains indices to form groups
      IE [0,1,0,1] returns [[0,2], [1,3]]
    """
    original = np.asanyarray(values)

    # save the sorted order and then apply it
    order = original.argsort()
    values = original[order]

    # find the indexes which are duplicates
    if values.dtype.kind == "f":
        # for floats in a sorted array, neighbors are not duplicates
        # if the difference between them is greater than approximate zero
        nondupe = np.greater(np.abs(np.diff(values)), tol.zero)
    else:
        # for ints and strings we can check exact non- equality
        # for all other types this will only work if they defined
        # an __eq__
        nondupe = values[1:] != values[:-1]

    dupe_idx = np.append(0, np.nonzero(nondupe)[0] + 1)

    # start with a mask that marks everything as ok
    dupe_ok = np.ones(len(dupe_idx), dtype=bool)

    # calculate the length of each group from their index
    dupe_len = np.diff(np.concatenate((dupe_idx, [len(values)])))

    # cull by length if requested
    if min_len is not None or max_len is not None:
        if min_len is not None:
            dupe_ok &= dupe_len >= min_len
        if max_len is not None:
            dupe_ok &= dupe_len <= max_len

    groups = [order[i : (i + j)] for i, j in zip(dupe_idx[dupe_ok], dupe_len[dupe_ok])]
    return groups


def hashable_rows(
    data: ArrayLike, digits: Optional[Integer] = None, allow_int: bool = True
) -> NDArray:
    """
    We turn our array into integers based on the precision
    given by digits and then put them in a hashable format.

    Parameters
    ---------
    data : (n, m) array
      Input data
    digits : int or None
      How many digits to add to hash if data is floating point
      If None, tol.merge will be used

    Returns
    ---------
    hashable : (n,)
      May return as a `np.void` or a `np.uint64`
    """
    # if there is no data return immediately
    if len(data) == 0:
        return np.array([], dtype=np.uint64)

    # get array as integer to precision we care about
    as_int = float_to_int(data, digits=digits)

    # if it is flat integers already return
    if len(as_int.shape) == 1:
        return as_int

    # if array is 2D and smallish, we can try bitbanging
    # this is significantly faster than the custom dtype
    if allow_int and len(as_int.shape) == 2 and as_int.shape[1] <= 4:
        # can we pack the whole row into a single 64 bit integer
        precision = int(np.floor(64 / as_int.shape[1]))

        # get the extreme values of the data set
        d_min, d_max = as_int.min(), as_int.max()
        # since we are quantizing the data down we need every value
        # to fit in a partial integer so we have to check against extrema
        threshold = (2 ** (precision - 1)) - 1

        # if the data is within the range of our precision threshold
        if d_max < threshold and d_min > -threshold:
            # the resulting package
            hashable = np.zeros(len(as_int), dtype=np.uint64)
            # offset to the middle of the unsigned integer range
            # this array should contain only positive values
            bitbang = (as_int.T + (threshold + 1)).astype(np.uint64)
            # loop through each column and bitwise xor to combine
            # make sure as_int is int64 otherwise bit offset won't work
            for offset, column in enumerate(bitbang):
                # will modify hashable in place
                np.bitwise_xor(hashable, column << (offset * precision), out=hashable)
            return hashable

    # reshape array into magical data type that is weird but works with unique
    dtype = np.dtype((np.void, as_int.dtype.itemsize * as_int.shape[1]))
    # make sure result is contiguous and flat
    result = np.ascontiguousarray(as_int).view(dtype).reshape(-1)
    result.flags["WRITEABLE"] = False

    return result


def float_to_int(data, digits: Optional[Integer] = None) -> NDArray[np.int64]:
    """
    Given a numpy array of float/bool/int, return as integers.

    Parameters
    -------------
    data :  (n, d) float, int, or bool
      Input data
    digits : float or int
      Precision for float conversion

    Returns
    -------------
    as_int : (n, d) int
      Data as integers
    """
    # convert to any numpy array
    data = np.asanyarray(data)

    # we can early-exit if we've been passed data that is already
    # an integer, unsigned integer, boolean, or empty
    if data.dtype == np.int64:
        return data
    elif data.dtype.kind in "iub" or data.size == 0:
        return data.astype(np.int64)
    elif data.dtype.kind != "f":
        # if it's not a floating point try to make it one
        data = data.astype(np.float64)

    if digits is None:
        # get digits from `tol.merge`
        digits = util.decimal_to_digits(tol.merge)
    elif not isinstance(digits, (int, np.integer)):
        raise TypeError(f"Digits must be `None` or `int`, not `{type(digits)}`")

    # multiply by requested power of ten
    # then subtract small epsilon to avoid "go either way" rounding
    # then do the rounding and convert to integer
    return np.round((data * 10**digits) - 1e-6).astype(np.int64)


def unique_ordered(
    data: ArrayLike, return_index: bool = False, return_inverse: bool = False
):
    """
    Returns the same as np.unique, but ordered as per the
    first occurrence of the unique value in data.

    Examples
    ---------
    In [1]: a = [0, 3, 3, 4, 1, 3, 0, 3, 2, 1]

    In [2]: np.unique(a)
    Out[2]: array([0, 1, 2, 3, 4])

    In [3]: trimesh.grouping.unique_ordered(a)
    Out[3]: array([0, 3, 4, 1, 2])
    """
    # uniques are the values, sorted
    # index is the value in the original `data`
    # i.e. `data[index] == unique`
    # inverse is how to re-construct `data` from `unique`
    # i.e. `unique[inverse] == data`
    unique, index, inverse = np.unique(data, return_index=True, return_inverse=True)

    # we want to maintain the original index order
    order = index.argsort()

    if not return_index and not return_inverse:
        return unique[order]

    # collect return values
    # start with the unique values in original order
    result = [unique[order]]
    # the new index values
    if return_index:
        # re-order the index in the original array
        result.append(index[order])
    if return_inverse:
        # create the new inverse from the order of the order
        result.append(order.argsort()[inverse])

    return result


def unique_bincount(
    values: ArrayLike,
    minlength: Integer = 0,
    return_inverse: bool = False,
    return_counts: bool = False,
):
    """
    For arrays of integers find unique values using bin counting.
    Roughly 10x faster for correct input than np.unique

    Parameters
    --------------
    values : (n,) int
      Values to find unique members of
    minlength : int
      Maximum value that will occur in values (values.max())
    return_inverse : bool
      If True, return an inverse such that unique[inverse] == values
    return_counts : bool
      If True, also return the number of times each
      unique item appears in values

    Returns
    ------------
    unique : (m,) int
      Unique values in original array
    inverse : (n,) int, optional
      An array such that unique[inverse] == values
      Only returned if return_inverse is True
    counts : (m,) int, optional
      An array holding the counts of each unique item in values
      Only returned if return_counts is True
    """
    values = np.asanyarray(values)
    if len(values.shape) != 1 or values.dtype.kind != "i":
        raise ValueError("input must be 1D integers!")

    try:
        # count the number of occurrences of each value
        counts = np.bincount(values, minlength=minlength)
    except TypeError:
        # casting failed on 32 bit windows
        log.warning("casting failed, falling back!")
        # fall back to numpy unique
        return np.unique(
            values, return_inverse=return_inverse, return_counts=return_counts
        )

    # which bins are occupied at all
    # counts are integers so this works
    unique_bin = counts.astype(bool)

    # which values are unique
    # indexes correspond to original values
    unique = np.where(unique_bin)[0]
    ret = (unique,)

    if return_inverse:
        # find the inverse to reconstruct original
        inverse = (np.cumsum(unique_bin) - 1)[values]
        ret += (inverse,)

    if return_counts:
        unique_counts = counts[unique]
        ret += (unique_counts,)

    if len(ret) == 1:
        return ret[0]
    return ret


def merge_runs(data: ArrayLike, digits: Optional[Integer] = None):
    """
    Merge duplicate sequential values. This differs from unique_ordered
    in that values can occur in multiple places in the sequence, but
    only consecutive repeats are removed

    Parameters
    -----------
    data: (n,) float or int

    Returns
    --------
    merged: (m,) float or int

    Examples
    ---------
    In [1]: a
    Out[1]:
    array([-1, -1, -1,  0,  0,  1,  1,  2,  0,
            3,  3,  4,  4,  5,  5,  6,  6,  7,
            7,  8,  8,  9,  9,  9])

    In [2]: trimesh.grouping.merge_runs(a)
    Out[2]: array([-1,  0,  1,  2,  0,  3,  4,  5,  6,  7,  8,  9])
    """
    if digits is None:
        epsilon = tol.merge
    else:
        epsilon = 10 ** (-digits)

    data = np.asanyarray(data)
    mask = np.zeros(len(data), dtype=bool)
    mask[0] = True
    mask[1:] = np.abs(data[1:] - data[:-1]) > epsilon

    return data[mask]


def unique_float(
    data,
    return_index: bool = False,
    return_inverse: bool = False,
    digits: Optional[Integer] = None,
):
    """
    Identical to the numpy.unique command, except evaluates floating point
    numbers, using a specified number of digits.

    If digits isn't specified, the library default TOL_MERGE will be used.
    """
    data = np.asanyarray(data)
    as_int = float_to_int(data, digits)
    _junk, unique, inverse = np.unique(as_int, return_index=True, return_inverse=True)

    if (not return_index) and (not return_inverse):
        return data[unique]

    result = [data[unique]]

    if return_index:
        result.append(unique)
    if return_inverse:
        result.append(inverse)
    return tuple(result)


def unique_rows(data, digits=None, keep_order=False):
    """
    Returns indices of unique rows. It will return the
    first occurrence of a row that is duplicated:
    [[1,2], [3,4], [1,2]] will return [0,1]

    Parameters
    ---------
    data : (n, m) array
      Floating point data
    digits : int or None
      How many digits to consider

    Returns
    --------
    unique :  (j,) int
      Index in data which is a unique row
    inverse : (n,) int
      Array to reconstruct original
      Example: data[unique][inverse] == data
    """
    # get rows hashable so we can run unique function on it
    rows = hashable_rows(data, digits=digits)

    # we are throwing away the first value which is the
    # garbage row-hash and only returning index and inverse
    if keep_order:
        # keeps order of original occurrence
        return unique_ordered(rows, return_index=True, return_inverse=True)[1:]
    # returns values sorted by row-hash but since our row-hash
    # were pretty much garbage the sort order isn't meaningful
    return np.unique(rows, return_index=True, return_inverse=True)[1:]


def unique_value_in_row(data, unique=None):
    """
    For a 2D array of integers find the position of a
    value in each row which only occurs once.

    If there are more than one value per row which
    occur once, the last one is returned.

    Parameters
    ----------
    data :   (n, d) int
      Data to check values
    unique : (m,) int
      List of unique values contained in data.
      Generated from np.unique if not passed

    Returns
    ---------
    result : (n, d) bool
      With one or zero True values per row.


    Examples
    -------------------------------------
    In [0]: r = np.array([[-1,  1,  1],
                          [-1,  1, -1],
                          [-1,  1,  1],
                          [-1,  1, -1],
                          [-1,  1, -1]], dtype=np.int8)

    In [1]: unique_value_in_row(r)
    Out[1]:
           array([[ True, False, False],
                  [False,  True, False],
                  [ True, False, False],
                  [False,  True, False],
                  [False,  True, False]], dtype=bool)

    In [2]: unique_value_in_row(r).sum(axis=1)
    Out[2]: array([1, 1, 1, 1, 1])

    In [3]: r[unique_value_in_row(r)]
    Out[3]: array([-1,  1, -1,  1,  1], dtype=int8)
    """
    if unique is None:
        unique = np.unique(data)
    data = np.asanyarray(data)
    result = np.zeros_like(data, dtype=bool, subok=False)
    for value in unique:
        test = np.equal(data, value)
        test_ok = test.sum(axis=1) == 1
        result[test_ok] = test[test_ok]
    return result


def group_rows(data, require_count=None, digits=None):
    """
    Returns index groups of duplicate rows, for example:
    [[1,2], [3,4], [1,2]] will return [[0,2], [1]]


    Note that using require_count allows numpy advanced
    indexing to be used in place of looping and
    checking hashes and is ~10x faster.


    Parameters
    ----------
    data : (n, m) array
      Data to group
    require_count : None or int
      Only return groups of a specified length, eg:
      require_count =  2
      [[1,2], [3,4], [1,2]] will return [[0,2]]
    digits : None or int
    If data is floating point how many decimals
    to consider, or calculated from tol.merge

    Returns
    ----------
    groups : sequence (*,) int
      Indices from in indicating identical rows.
    """

    # start with getting a sortable format
    hashable = hashable_rows(data, digits=digits)

    # if there isn't a constant column size use more complex logic
    if require_count is None:
        return group(hashable)

    # record the order of the rows so we can get the original indices back
    order = hashable.argsort()
    # but for now, we want our hashes sorted
    hashable = hashable[order]
    # this is checking each neighbour for equality, example:
    # example: hashable = [1, 1, 1]; dupe = [0, 0]
    dupe = hashable[1:] != hashable[:-1]
    # we want the first index of a group, so we can slice from that location
    # example: hashable = [0 1 1]; dupe = [1,0]; dupe_idx = [0,1]
    dupe_idx = np.append(0, np.nonzero(dupe)[0] + 1)
    # if you wanted to use this one function to deal with non- regular groups
    # you could use: np.array_split(dupe_idx)
    # this is roughly 3x slower than using the group_dict method above.
    start_ok = np.diff(np.concatenate((dupe_idx, [len(hashable)]))) == require_count
    groups = np.tile(dupe_idx[start_ok].reshape((-1, 1)), require_count) + np.arange(
        require_count
    )
    groups_idx = order[groups]

    if require_count == 1:
        return groups_idx.reshape(-1)
    return groups_idx


def boolean_rows(
    a: ArrayLike, b: ArrayLike, operation=np.intersect1d
) -> NDArray[np.int64]:
    """
    Find the rows in two arrays which occur in both rows.

    Parameters
    ---------
    a: (n, d) int
        Array with row vectors
    b: (m, d) int
        Array with row vectors
    operation : function
        Numpy boolean set operation function:
          -np.intersect1d
          -np.setdiff1d

    Returns
    --------
    shared : (p, d) int64
       Array containing requested rows in both a and b
    """
    a = np.asanyarray(a, dtype=np.int64)
    b = np.asanyarray(b, dtype=np.int64)

    av = a.view([("", a.dtype)] * a.shape[1]).ravel()
    bv = b.view([("", b.dtype)] * b.shape[1]).ravel()
    return operation(av, bv).view(a.dtype).reshape(-1, a.shape[1])


def group_vectors(vectors, angle=1e-4, include_negative=False):
    """
    Group vectors based on an angle tolerance, with the option to
    include negative vectors.

    Parameters
    -----------
    vectors : (n,3) float
        Direction vector
    angle : float
        Group vectors closer than this angle in radians
    include_negative : bool
        If True consider the same:
        [0,0,1] and [0,0,-1]

    Returns
    ------------
    new_vectors : (m,3) float
        Direction vector
    groups : (m,) sequence of int
        Indices of source vectors
    """

    vectors = np.asanyarray(vectors, dtype=np.float64)
    angle = float(angle)

    if include_negative:
        vectors = util.vector_hemisphere(vectors)

    spherical = util.vector_to_spherical(vectors)
    angles, groups = group_distance(spherical, angle)
    new_vectors = util.spherical_to_vector(angles)
    return new_vectors, groups


def group_distance(
    values: ArrayLike, distance: Number
) -> Tuple[NDArray[np.float64], Sequence]:
    """
    Find non-overlapping groups of points where no two points in a
    group are farther than 2*distance apart.

    Parameters
    ---------
    values : (n, d) float
        Points of dimension d
    distance : float
        Max distance between points in a cluster

    Returns
    ----------
    unique : (m, d) float
        Median value of each group
    groups : (m) sequence of int
        Indexes of points that make up a group

    """
    values = np.asanyarray(values, dtype=np.float64)

    consumed = np.zeros(len(values), dtype=bool)
    tree = cKDTree(values)

    # (n, d) set of values that are unique
    unique = []
    # (n) sequence of indices in values
    groups = []

    for index, value in enumerate(values):
        if consumed[index]:
            continue
        group = np.array(tree.query_ball_point(value, distance), dtype=np.int64)
        group = group[~consumed[group]]
        consumed[group] = True
        unique.append(np.median(values[group], axis=0))
        groups.append(group)
    return np.array(unique), groups


def clusters(points, radius):
    """
    Find clusters of points which have neighbours closer than radius

    Parameters
    ---------
    points : (n, d) float
        Points of dimension d
    radius : float
        Max distance between points in a cluster

    Returns
    ----------
    groups : (m,) sequence of int
        Indices of points in a cluster

    """
    from . import graph

    tree = cKDTree(points)

    # some versions return pairs as a set of tuples
    pairs = tree.query_pairs(r=radius, output_type="ndarray")
    # group connected components
    groups = graph.connected_components(pairs)

    return groups


def blocks(data, min_len=2, max_len=np.inf, wrap=False, digits=None, only_nonzero=False):
    """
    Find the indices in an array of contiguous blocks
    of equal values.

    Parameters
    ------------
    data : (n,) array
      Data to find blocks on
    min_len : int
      The minimum length group to be returned
    max_len : int
      The maximum length group to be retuurned
    wrap : bool
      Combine blocks on both ends of 1D array
    digits : None or int
      If dealing with floats how many digits to consider
    only_nonzero : bool
      Only return blocks of non- zero values

    Returns
    ---------
    blocks : (m) sequence of (*,) int
      Indices referencing data
    """
    data = float_to_int(data, digits=digits)

    # keep an integer range around so we can slice
    arange = np.arange(len(data))
    arange.flags["WRITEABLE"] = False

    nonzero = arange[1:][data[1:] != data[:-1]]
    infl = np.zeros(len(nonzero) + 2, dtype=int)
    infl[-1] = len(data)
    infl[1:-1] = nonzero

    # the length of each chunk
    infl_len = infl[1:] - infl[:-1]

    # check the length of each group
    infl_ok = np.logical_and(infl_len >= min_len, infl_len <= max_len)

    if only_nonzero:
        # check to make sure the values of each contiguous block
        # are True by checking the first value of each block
        infl_ok = np.logical_and(infl_ok, data[infl[:-1]])

    # inflate start/end indexes into full ranges of values
    blocks = [arange[infl[i] : infl[i + 1]] for i, ok in enumerate(infl_ok) if ok]

    if wrap:
        # wrap only matters if first and last points are the same
        if data[0] != data[-1]:
            return blocks
        # if we are only grouping nonzero things and
        # the first and last point are zero we can exit
        if only_nonzero and not bool(data[0]):
            return blocks

        # if all values are True or False we can exit
        if len(blocks) == 1 and len(blocks[0]) == len(data):
            return blocks

        # so now first point equals last point, so the cases are:
        # - first and last point are in a block: combine two blocks
        # - first OR last point are in block: add other point to block
        # - neither are in a block: check if combined is eligible block

        # first point is in a block
        first = len(blocks) > 0 and blocks[0][0] == 0
        # last point is in a block
        last = len(blocks) > 0 and blocks[-1][-1] == (len(data) - 1)

        # CASE: first and last point are BOTH in block: combine blocks
        if first and last:
            blocks[0] = np.append(blocks[-1], blocks[0])
            blocks.pop()
        else:
            # combined length
            combined = infl_len[0] + infl_len[-1]
            # exit if lengths aren't OK
            if combined < min_len or combined > max_len:
                return blocks
            # new block combines both ends
            new_block = np.append(
                np.arange(infl[-2], infl[-1]), np.arange(infl[0], infl[1])
            )
            # we are in a first OR last situation now
            if first:
                # first was already in a block so replace it with combined
                blocks[0] = new_block
            elif last:
                # last was already in a block so replace with superset
                blocks[-1] = new_block
            else:
                # both are false
                # combined length generated new block
                blocks.append(new_block)

    return blocks


def group_min(groups, data):
    """
    Given a list of groups find the minimum element of data
    within each group

    Parameters
    -----------
    groups : (n,) sequence of (q,) int
        Indexes of each group corresponding to each element in data
    data : (m,)
        The data that groups indexes reference

    Returns
    -----------
    minimums : (n,)
        Minimum value of data per group

    """
    # sort with major key groups, minor key data
    order = np.lexsort((data, groups))
    groups = groups[order]  # this is only needed if groups is unsorted
    data = data[order]
    # construct an index which marks borders between groups
    index = np.zeros(len(groups), "bool")
    index[0] = True
    index[1:] = groups[1:] != groups[:-1]
    return data[index]



try:
    from . import generic as g
except BaseException:
    import generic as g


class MetaTest(g.unittest.TestCase):
    def test_glb(self):
        # check to see if both `Scene.metadata` and
        # `Scene.geometry._.metadata` is preserved

        # create an example scene
        s = g.trimesh.Scene(
            [
                g.trimesh.creation.box().permutate.transform(),
                g.trimesh.creation.box().permutate.transform(),
                g.trimesh.creation.box().permutate.transform(),
            ]
        )

        # add some basic metadata
        s.metadata["hi"] = True
        s.metadata["10"] = "it's true!"
        for m in s.geometry.values():
            # create some random metadata for each mesh
            # note that JSON doesn't support integers as keys
            # so convert integers to strings for comparison
            m.metadata.update(
                g.np.random.randint(0, 1000, 10).reshape((-1, 2)).astype(str)
            )

        # reload the exported scene
        r = g.trimesh.load(
            file_obj=g.trimesh.util.wrap_as_stream(s.export(file_type="glb")),
            file_type="glb",
        )
        # all scene metadata should have survived export-import cycle
        assert r.metadata == s.metadata

        # all geometry should have the same names
        assert set(r.geometry.keys()) == set(s.geometry.keys())

        for k, a in s.geometry.items():
            # `a` is the original mesh
            # `b` is the same mesh after export-import
            b = r.geometry[k]

            # the original metadata should have all survived
            # the exporter is allowed to add additional keys
            assert set(b.metadata.keys()).issuperset(a.metadata.keys())

            # every original value must match exactly
            assert all(b.metadata[k] == v for k, v in a.metadata.items())

    def test_svg(self):
        p = g.get_mesh("2D/1002_tray_bottom.DXF")

        assert len(p.layers) == len(p.entities)
        assert all(e.layer == "0" for e in p.entities)
        assert all(L == "0" for L in p.layers)

        r = g.trimesh.load(
            file_obj=g.trimesh.util.wrap_as_stream(p.export(file_type="svg")),
            file_type="svg",
        )

        # make sure we didn't stomp on original
        assert all(e.layer == "0" for e in p.entities)
        assert all(e.layer == "0" for e in r.entities)
        assert all(i == j for i, j in zip(p.layers, r.layers))
        assert p.metadata == r.metadata
        assert len(p.metadata) > 0


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class ResolverTest(g.unittest.TestCase):
    def test_filepath_namespace(self):
        # check the namespaced method
        models = g.dir_models
        subdir = "2D"

        # create a resolver for the models directory
        resolver = g.trimesh.resolvers.FilePathResolver(models)

        # should be able to get an asset
        assert len(resolver.get("rabbit.obj")) > 0

        # check a few file path keys
        check = {"ballA.off", "featuretype.STL"}
        assert set(resolver.keys()).issuperset(check)

        # try a namespaced resolver
        ns = resolver.namespaced(subdir)
        assert not set(ns.keys()).issuperset(check)
        assert set(ns.keys()).issuperset(["tray-easy1.dxf", "single_arc.dxf"])

    def test_web_namespace(self):
        base = "https://example.com"
        name = "stuff"
        target = "hi.gltf"

        # check with a trailing slash
        a = g.trimesh.resolvers.WebResolver(base + "/")
        b = g.trimesh.resolvers.WebResolver(base + "//")
        c = g.trimesh.resolvers.WebResolver(base)
        d = a.namespaced(name)

        # base URL's should always be the same with one trailing slash
        assert a.base_url == b.base_url
        assert b.base_url == c.base_url
        assert c.base_url == base + "/"
        # check namespaced
        assert d.base_url == base + "/" + name + "/"

        # should have correct slashes
        truth = "/".join([base, name, target])

        assert a.base_url + name + "/" + target == truth
        assert d.base_url + target == truth

    def test_items(self):
        # check __getitem__ and __setitem__
        archive = {}
        resolver = g.trimesh.resolvers.ZipResolver(archive)
        assert len(set(resolver.keys())) == 0
        resolver["hi"] = b"what"
        # should have one item
        assert set(resolver.keys()) == {"hi"}
        # should have the right value
        assert resolver["hi"] == b"what"
        # original archive should have been modified
        assert set(archive.keys()) == {"hi"}

        # add a subdirectory key
        resolver["stuff/nah"] = b"sup"
        assert set(archive.keys()) == {"hi", "stuff/nah"}
        assert set(resolver.keys()) == {"hi", "stuff/nah"}

        # try namespacing
        ns = resolver.namespaced("stuff")
        assert ns["nah"] == b"sup"
        g.log.debug(ns.keys())
        assert set(ns.keys()) == {"nah"}


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



import numpy as np
from pyinstrument import Profiler
from shapely.geometry import Point, Polygon

import trimesh
from trimesh.creation import sweep_polygon

trimesh.tol.strict = True


def arc_2d(final_angle, d=1, start_angle=0, splits=20):
    xs = np.linspace(start_angle, final_angle, splits)
    return np.stack((d / 2 * np.cos(xs), d / 2 * np.sin(xs)), axis=1)


def test_simple_closed(h1=1, w=2, r=4.4):
    # a simple closed path
    square = Polygon([(-0.01, -h1 / 2), (w, -h1 / 2), (w, +h1 / 2), (-0.01, +h1 / 2)])
    circle = Point([0, 0]).buffer(1.0)

    # a simple square of side `r`
    path = [[0, 0, 0], [r, 0, 0], [r, r, 0], [0, r, 0], [0, 0, 0]]

    a = sweep_polygon(square, path)
    assert a.is_volume

    aa = sweep_polygon(square, path, connect=False)
    assert aa.is_volume

    # should have the same bounds but no longer a volume
    aaa = sweep_polygon(square, path, connect=False, cap=False)
    assert np.allclose(aa.bounds, aaa.bounds)
    assert not aaa.is_volume

    b = sweep_polygon(circle, path)
    assert b.is_volume


def test_simple_extrude(height=10):
    # make sure sweep behaves okay on a simple single segment path
    # this should be identical to the extrude operation
    circle = Point([0, 0]).buffer(1.0)

    # a simple extrusion
    path = [[0, 0, 0], [0, 0, height]]
    # will be running asserts inside function
    b = sweep_polygon(circle, path)

    # should be a straight extrude along Z
    expected = np.append(np.ptp(np.reshape(circle.bounds, (2, 2)), axis=0), height)
    assert np.allclose(expected, b.extents)

    # should be well constructed
    assert b.is_volume
    # volume should correspond to expected cylinder area
    assert np.isclose(b.volume, circle.area * height)


def test_simple_open():
    circle = Point([0, 0]).buffer(1.0)
    theta = np.linspace(0.0, np.pi, 100)
    path = np.column_stack([np.cos(theta), np.sin(theta), np.zeros(len(theta))]) * 10

    a = sweep_polygon(circle, path)
    assert a.is_volume


def test_spline_3D():
    circle = Point([0, 0]).buffer(1.0)
    # make a 3D noodle using spline smoothing
    path = trimesh.path.simplify.resample_spline(
        [[0, 0, 0], [4, 4, 0], [5, 5, 0], [10, 0, 10], [0, 20, 0]], smooth=0.25, count=100
    )

    a = sweep_polygon(circle, path)
    assert a.is_volume
    assert a.body_count == 1


def test_screw():
    h = 200
    d = 8
    h1 = 1
    w = 2
    d = d - w
    h2 = h1

    lead = 2
    splits_per_turn = 4 * 20
    polygon = np.array([(-0.01, -h1 / 2), (w, -h2 / 2), (w, +h2 / 2), (-0.01, +h1 / 2)])
    h_ = h + h1
    n = int(np.ceil(splits_per_turn * h_ / lead))
    turns = n / splits_per_turn
    xys = arc_2d(2 * np.pi * turns, d, splits=n)
    zs = np.linspace(0, h_, n)[:, None]
    spin_path = np.concatenate((xys, zs), axis=1)

    with Profiler() as P:
        x = sweep_polygon(Polygon(polygon), spin_path)
    P.print()

    assert x.is_volume
    # should have produced a result corresponding to the input dimensions
    assert np.allclose(x.extents[:2], d + w * 2, atol=0.01)
    assert np.allclose(x.extents[2], h + h1 * 2, atol=0.01)

    # check without capping: should be the same size just without caps
    o = sweep_polygon(Polygon(polygon), spin_path, cap=False, connect=False)
    assert np.allclose(x.bounds, o.bounds)
    assert not o.is_volume

    # try spinning along an angle
    sweep_polygon(
        Polygon(polygon), spin_path, angles=np.linspace(0.0, np.pi / 2, len(spin_path))
    )

    circle = Point([0, 0]).buffer(1.0)
    theta = np.linspace(0.0, np.pi, 100)
    path = np.column_stack([np.cos(theta), np.sin(theta), np.zeros(len(theta))]) * 10

    a = sweep_polygon(circle, path)
    assert a.is_volume


if __name__ == "__main__":
    test_simple_closed()
    test_simple_extrude()
    test_simple_open()
    test_screw()
    test_spline_3D()



import collections
from copy import deepcopy

import numpy as np

from .. import caching, util
from ..caching import hash_fast
from ..transformations import fix_rigid, quaternion_matrix, rotation_matrix
from ..typed import ArrayLike, Hashable, NDArray, Optional, Sequence, Tuple, Union

# we compare to identity a lot
_identity = np.eye(4)
_identity.flags["WRITEABLE"] = False


class SceneGraph:
    """
    Hold data about positions and instances of geometry
    in a scene. This includes a forest (i.e. multi-root tree)
    of transforms and information on which node is the base
    frame, and which geometries are affiliated with which
    nodes.
    """

    def __init__(self, base_frame="world", repair_rigid=1e-5):
        """
        Create a scene graph, holding homogeneous transformation
        matrices and instance information about geometry.

        Parameters
        -----------
        base_frame : any
          The root node transforms will be positioned from.
        repair_rigid : None or float
          If a float will attempt to repair rotation matrices
          where `M @ M.T` differs from an identity matrix by
          more than floating point zero but less than this value.
          This can happen in a deep tree with a lot of matrix
          multiplies.
        """
        # a graph structure, subclass of networkx DiGraph
        self.transforms = EnforcedForest()
        # hashable, the base or root frame
        self.base_frame = base_frame
        # if passed as a float try to repair rigid transforms
        # that have accumulated floating point error
        self.repair_rigid = repair_rigid
        # cache transformation matrices keyed with tuples
        self._cache = caching.Cache(self.__hash__)

    def update(self, frame_to, frame_from=None, **kwargs):
        """
        Update a transform in the tree.

        Parameters
        ------------
        frame_from : hashable object
          Usually a string (eg 'world').
          If left as None it will be set to self.base_frame
        frame_to :  hashable object
          Usually a string (eg 'mesh_0')
        matrix : (4,4) float
          Homogeneous transformation matrix
        quaternion :  (4,) float
          Quaternion ordered [w, x, y, z]
        axis : (3,) float
          Axis of rotation
        angle :  float
          Angle of rotation, in radians
        translation : (3,) float
          Distance to translate
        geometry : hashable
          Geometry object name, e.g. 'mesh_0'
        metadata: dictionary
          Optional metadata attached to the new frame
          (exports to glTF node 'extras').
        """
        # if no frame specified, use base frame
        if frame_from is None:
            frame_from = self.base_frame

        # pass through
        attr = {k: v for k, v in kwargs.items() if k in {"geometry", "metadata"}}
        # convert various kwargs to a single matrix
        attr["matrix"] = kwargs_to_matrix(**kwargs)

        # add the edges for the transforms
        # wi ll return if it changed anything
        self.transforms.add_edge(frame_from, frame_to, **attr)

        # set the node attribute with the geometry information
        if "geometry" in kwargs:
            self.transforms.node_data[frame_to]["geometry"] = kwargs["geometry"]

    def get(
        self, frame_to: Hashable, frame_from: Optional[Hashable] = None
    ) -> Tuple[NDArray[np.float64], Optional[Hashable]]:
        """
        Get the transform from one frame to another.

        Parameters
        ------------
        frame_to : hashable
          Node name, usually a string (eg 'mesh_0')
        frame_from : hashable
          Node name, usually a string (eg 'world').
          If None it will be set to self.base_frame

        Returns
        ----------
        transform : (4, 4) float
          Homogeneous transformation matrix
        geometry
          The name of the geometry if it exists

        Raises
        -----------
        ValueError
          If the frames aren't connected.
        """

        # use base frame if not specified
        if frame_from is None:
            frame_from = self.base_frame

        # look up transform to see if we have it already
        key = (frame_from, frame_to)
        if key in self._cache:
            return self._cache[key]

        # get the geometry at the final node if any
        geometry = self.transforms.node_data[frame_to].get("geometry")

        # get a local reference to edge data
        data = self.transforms.edge_data

        if frame_from == frame_to:
            # if we're going from ourself return identity
            matrix = _identity
        elif key in data:
            # if the path is just an edge return early
            matrix = data[key]["matrix"]
        else:
            # we have a 3+ node path
            # get the path from the forest always going from
            # parent -> child -> child
            path = self.transforms.shortest_path(frame_from, frame_to)
            # the path should always start with `frame_from`
            assert path[0] == frame_from
            # and end with the `frame_to` node
            assert path[-1] == frame_to

            # loop through pairs of the path
            matrices = []
            for u, v in zip(path[:-1], path[1:]):
                forward = data.get((u, v))
                if forward is not None:
                    if "matrix" in forward:
                        # append the matrix from u to v
                        matrices.append(forward["matrix"])
                    continue
                # since forwards didn't exist backward must
                # exist otherwise this is a disconnected path
                # and we should raise an error anyway
                backward = data[(v, u)]
                if "matrix" in backward:
                    # append the inverted backwards matrix
                    matrices.append(np.linalg.inv(backward["matrix"]))
            # filter out any identity matrices
            matrices = [m for m in matrices if np.abs(m - _identity).max() > 1e-8]
            if len(matrices) == 0:
                matrix = _identity
            elif len(matrices) == 1:
                matrix = matrices[0]
            else:
                # multiply matrices into single transform
                matrix = util.multi_dot(matrices)

        # if instructed to repair rigid transforms do it here
        if self.repair_rigid is not None:
            matrix = fix_rigid(matrix, max_deviance=self.repair_rigid)

        # matrix being edited in-place leads to subtle bugs
        matrix.flags["WRITEABLE"] = False

        # store the result
        self._cache[key] = (matrix, geometry)

        return matrix, geometry

    def __hash__(self):
        return self.transforms.__hash__()

    def copy(self):
        """
        Return a copy of the current TransformForest.

        Returns
        ------------
        copied : TransformForest
          Copy of current object.
        """
        # create a copy without transferring cache
        copied = SceneGraph()
        copied.base_frame = deepcopy(self.base_frame)
        copied.transforms = deepcopy(self.transforms)
        return copied

    def to_flattened(self):
        """
        Export the current transform graph with all
        transforms baked into world->instance.

        Returns
        ---------
        flat : dict
          Keyed {node : {transform, geometry}
        """
        flat = {}
        base_frame = self.base_frame
        for node in self.nodes:
            if node == base_frame:
                continue
            # get the matrix and geometry name
            matrix, geometry = self.get(frame_to=node, frame_from=base_frame)
            # store matrix as list rather than numpy array
            flat[node] = {"transform": matrix.tolist(), "geometry": geometry}

        return flat

    def to_gltf(self, scene, mesh_index=None):
        """
        Export a transforms as the 'nodes' section of the
        GLTF header dict.

        Parameters
        ------------
        scene : trimesh.Scene
          Scene with geometry.
        mesh_index : dict or None
          Mapping { key in scene.geometry : int }

        Returns
        --------
        gltf : dict
          With 'nodes' referencing a list of dicts
        """

        if mesh_index is None:
            # geometry is an OrderedDict
            # map mesh name to index: {geometry key : index}
            mesh_index = {name: i for i, name in enumerate(scene.geometry.keys())}

        # get graph information into local scope before loop
        graph = self.transforms
        # get the stored node data
        node_data = graph.node_data
        edge_data = graph.edge_data
        base_frame = self.base_frame

        # list of dict, in gltf format
        # start with base frame as first node index
        result = [{"name": base_frame}]
        # {node name : node index in gltf}
        lookup = {base_frame: 0}

        # collect the nodes in order
        for node in node_data.keys():
            if node == base_frame:
                continue
            # assign the index to the node-name lookup
            lookup[node] = len(result)
            # populate a result at the correct index
            result.append({"name": node})

        # get generated properties outside of loop
        # does the scene have a defined camera to export
        has_camera = scene.has_camera
        children = graph.children

        extensions_used = set()

        # then iterate through to collect data
        for info in result:
            # name of the scene node
            node = info["name"]

            # get the original node names for children
            childs = children.get(node, [])
            if len(childs) > 0:
                info["children"] = [lookup[k] for k in childs]

            # if we have a mesh store by index
            if "geometry" in node_data[node]:
                mesh_key = node_data[node]["geometry"]
                if mesh_key in mesh_index:
                    info["mesh"] = mesh_index[mesh_key]
            # check to see if we have camera node
            if has_camera and node == scene.camera.name:
                info["camera"] = 0

            if node != base_frame:
                parent = graph.parents[node]
                node_edge = edge_data[(parent, node)]

                # get the matrix from this edge
                matrix = node_edge["matrix"]
                # only include if it's not an identify matrix
                if not util.allclose(matrix, _identity):
                    info["matrix"] = matrix.T.reshape(-1).tolist()

                # if an extra was stored on this edge
                extras = node_edge.get("metadata")
                if extras:
                    extras = extras.copy()

                    # if extensionss were stored on this edge
                    extensions = extras.pop("gltf_extensions", None)
                    if isinstance(extensions, dict):
                        info["extensions"] = extensions
                        extensions_used = extensions_used.union(set(extensions.keys()))

                    # convert any numpy arrays to lists
                    extras.update(
                        {k: v.tolist() for k, v in extras.items() if hasattr(v, "tolist")}
                    )
                    info["extras"] = extras

        gltf = {"nodes": result}
        if len(extensions_used) > 0:
            gltf["extensionsUsed"] = list(extensions_used)
        return gltf

    def to_edgelist(self):
        """
        Export the current transforms as a list of
        edge tuples, with each tuple having the format:
        (node_a, node_b, {metadata})

        Returns
        ---------
        edgelist : (n,) list
          Of edge tuples
        """
        # save local reference to node_data
        nodes = self.transforms.node_data
        # save cleaned edges
        export = []
        # loop through (node, node, edge attributes)
        for edge, attr in self.transforms.edge_data.items():
            # node indexes from edge
            a, b = edge
            # geometry is a node property but save it to the
            # edge so we don't need two dictionaries
            b_attr = nodes[b]
            # make sure we're not stomping on original
            attr_new = attr.copy()
            # apply node geometry to edge attributes
            if "geometry" in b_attr:
                attr_new["geometry"] = b_attr["geometry"]
            # convert any numpy arrays to regular lists
            attr_new.update(
                {k: v.tolist() for k, v in attr_new.items() if hasattr(v, "tolist")}
            )
            export.append([a, b, attr_new])
        return export

    def from_edgelist(self, edges, strict=True):
        """
        Load transform data from an edge list into the current
        scene graph.

        Parameters
        -------------
        edgelist : (n,) tuples
          Keyed (node_a, node_b, {key: value})
        strict : bool
          If True raise a ValueError when a
          malformed edge is passed in a tuple.
        """

        # loop through each edge
        for edge in edges:
            # edge contains attributes
            if len(edge) == 3:
                self.update(edge[1], edge[0], **edge[2])
            # edge just contains nodes
            elif len(edge) == 2:
                self.update(edge[1], edge[0])
            # edge is broken
            elif strict:
                raise ValueError("edge incorrect shape: %s", str(edge))

    def to_networkx(self):
        """
        Return a `networkx` copy of this graph.

        Returns
        ----------
        graph : networkx.DiGraph
          Directed graph.
        """
        import networkx

        return networkx.from_edgelist(self.to_edgelist(), create_using=networkx.DiGraph)

    def show(self, **kwargs):
        """
        Plot the scene graph using `networkx.draw_networkx`
        which uses matplotlib to display the graph.

        Parameters
        -----------
        kwargs : dict
          Passed to `networkx.draw_networkx`
        """
        import matplotlib.pyplot as plt  # noqa
        import networkx

        # default kwargs will only be set if not
        # passed explicitly to the show command
        defaults = {"with_labels": True}
        kwargs.update(**{k: v for k, v in defaults.items() if k not in kwargs})
        networkx.draw_networkx(G=self.to_networkx(), **kwargs)

        plt.show()

    def load(self, edgelist):
        """
        Load transform data from an edge list into the current
        scene graph.

        Parameters
        -------------
        edgelist : (n,) tuples
          Structured (node_a, node_b, {key: value})
        """
        self.from_edgelist(edgelist, strict=True)

    @caching.cache_decorator
    def nodes(self):
        """
        A list of every node in the graph.

        Returns
        -------------
        nodes : (n,) array
          All node names.
        """
        return self.transforms.nodes

    @caching.cache_decorator
    def nodes_geometry(self):
        """
        The nodes in the scene graph with geometry attached.

        Returns
        ------------
        nodes_geometry : (m,) array
          Node names which have geometry associated
        """
        return [n for n, attr in self.transforms.node_data.items() if "geometry" in attr]

    @caching.cache_decorator
    def geometry_nodes(self):
        """
        Which nodes have this geometry? Inverse
        of `nodes_geometry`.

        Returns
        ------------
        geometry_nodes : dict
          Keyed {geometry_name : node name}
        """
        res = collections.defaultdict(list)
        for node, attr in self.transforms.node_data.items():
            if "geometry" in attr:
                res[attr["geometry"]].append(node)
        return res

    def remove_geometries(self, geometries: Union[str, set, Sequence]):
        """
        Remove the reference for specified geometries
        from nodes without deleting the node.

        Parameters
        ------------
        geometries : list or str
          Name of scene.geometry to dereference.
        """
        # make sure we have a set of geometries to remove
        if isinstance(geometries, str):
            geometries = [geometries]
        geometries = set(geometries)

        # remove the geometry reference from the node without deleting nodes
        # this lets us keep our cached paths, and will not screw up children
        for attrib in self.transforms.node_data.values():
            if "geometry" in attrib and attrib["geometry"] in geometries:
                attrib.pop("geometry")

        # it would be safer to just run _cache.clear
        # but the only property using the geometry should be
        # nodes_geometry: if this becomes not true change this to clear!
        self._cache.cache.pop("nodes_geometry", None)
        self.transforms._hash = None

    def __contains__(self, key: Hashable) -> bool:
        return key in self.transforms.node_data

    def __getitem__(
        self, key: Hashable
    ) -> Tuple[NDArray[np.float64], Optional[Hashable]]:
        return self.get(key)

    def __setitem__(self, key: Hashable, value: ArrayLike):
        value = np.asanyarray(value, dtype=np.float64)
        if value.shape != (4, 4):
            raise ValueError("Matrix must be specified!")
        return self.update(key, matrix=value)

    def clear(self):
        self.transforms = EnforcedForest()
        self._cache.clear()


class EnforcedForest:
    """
    A simple forest graph data structure: every node
    is allowed to have exactly one parent. This makes
    traversal and implementation much simpler than a
    full graph data type; by storing only one parent
    reference, it enforces the structure for "free."
    """

    def __init__(self):
        # since every node can have only one parent
        # this data structure transparently enforces
        # the forest data structure without checks
        # a dict {child : parent}
        self.parents = {}

        # store data for a particular edge keyed by tuple
        # {(u, v) : data }
        self.edge_data = collections.defaultdict(dict)
        # {u: data}
        self.node_data = collections.defaultdict(dict)

        # if multiple calls are made for the same path
        # but the connectivity hasn't changed return cached
        self._cache = {}

    def add_edge(self, u, v, **kwargs):
        """
        Add an edge to the forest cleanly.

        Parameters
        -----------
        u : any
          Hashable node key.
        v : any
          Hashable node key.
        kwargs : dict
           Stored as (u, v) edge data.

        Returns
        --------
        changed : bool
          Return if this operation changed anything.
        """
        self._hash = None

        # topology has changed so clear cache
        if (u, v) not in self.edge_data:
            self._cache = {}
        else:
            # check to see if matrix and geometry are identical
            edge = self.edge_data[(u, v)]
            if util.allclose(
                kwargs.get("matrix", _identity), edge.get("matrix", _identity), 1e-8
            ) and (edge.get("geometry") == kwargs.get("geometry")):
                return False

        # store a parent reference for traversal
        self.parents[v] = u
        # store kwargs for edge data keyed with tuple
        self.edge_data[(u, v)] = kwargs
        # set empty node data
        self.node_data[u].update({})
        if "geometry" in kwargs:
            self.node_data[v].update({"geometry": kwargs["geometry"]})
        else:
            self.node_data[v].update({})

        return True

    def remove_node(self, u):
        """
        Remove a node from the forest.

        Parameters
        -----------
        u : any
          Hashable node key.

        Returns
        --------
        changed : bool
          Return if this operation changed anything.
        """
        # check if node is part of forest
        if u not in self.node_data:
            return False

        # topology will change so clear cache
        self._cache = {}
        self._hash = None

        # delete all children's references and parent reference
        children = [child for (child, parent) in self.parents.items() if parent == u]
        for c in children:
            del self.parents[c]
        if u in self.parents:
            del self.parents[u]

        # delete edge data
        edges = [(a, b) for (a, b) in self.edge_data if a == u or b == u]
        for e in edges:
            del self.edge_data[e]

        # delete node data
        del self.node_data[u]

        return True

    def shortest_path(self, u, v):
        """
        Find the shortest path between `u` and `v`, returning
        a path where the first element is always `u` and the
        last element is always `v`, disregarding edge direction.

        Parameters
        -----------
        u : any
          Hashable node key.
        v : any
          Hashable node key.

        Returns
        -----------
        path : (n,)
          Path between `u` and `v`
        """
        # see if we've already computed this path
        if u == v:
            #  the path between itself is an edge case
            return []
        elif (u, v) in self._cache:
            # return the same path for either direction
            return self._cache[(u, v)]
        elif (v, u) in self._cache:
            return self._cache[(v, u)][::-1]

        # local reference to parent dict for performance
        parents = self.parents
        # store both forward and backwards traversal
        forward = [u]
        backward = [v]

        # cap iteration to number of total nodes
        for _ in range(len(parents) + 1):
            # store the parent both forwards and backwards
            f = parents.get(forward[-1])
            b = parents.get(backward[-1])
            forward.append(f)
            backward.append(b)

            if f == v:
                self._cache[(u, v)] = forward
                return forward
            elif b == u:
                # return reversed path
                backward = backward[::-1]
                self._cache[(u, v)] = backward
                return backward
            elif (b in forward) or (f is None and b is None):
                # we have a either a common node between both
                # traversal directions or we have consumed the whole
                # tree in both directions so try to find the common node
                common = set(backward).intersection(forward).difference({None})
                if len(common) == 0:
                    raise ValueError(f"No path from {u}->{v}!")
                elif len(common) > 1:
                    # get the first occurring common element in "forward"
                    link = next(f for f in forward if f in common)
                    assert link in common
                else:
                    # take the only common element
                    link = next(iter(common))

                # combine the forward and backwards traversals
                a = forward[: forward.index(link) + 1]
                b = backward[: backward.index(link)]
                path = a + b[::-1]

                # verify we didn't screw up the order
                assert path[0] == u
                assert path[-1] == v

                self._cache[(u, v)] = path

                return path

        raise ValueError("Iteration limit exceeded!")

    @property
    def nodes(self):
        """
        Get a set of every node.

        Returns
        -----------
        nodes : set
          Every node currently stored.
        """
        return self.node_data.keys()

    @property
    def children(self):
        """
        Get the children of each node.

        Returns
        ----------
        children : dict
          Keyed {node : [child, child, ...]}
        """
        if "children" in self._cache:
            return self._cache["children"]
        child = collections.defaultdict(list)
        # append children to parent references
        # skip self-references to avoid a node loop
        [child[v].append(u) for u, v in self.parents.items() if u != v]

        # cache and return as a vanilla dict
        self._cache["children"] = dict(child)
        return self._cache["children"]

    def successors(self, node):
        """
        Get all nodes that are successors to specified node,
        including the specified node.

        Parameters
        -------------
        node : any
          Hashable key for a node.

        Returns
        ------------
        successors : set
          Nodes that succeed specified node.
        """
        # get mapping of {parent : child}
        children = self.children
        # if node doesn't exist return early
        if node not in children:
            return {node}

        # children we need to collect
        queue = [node]
        # start collecting values with children of source
        collected = set(queue)

        # cap maximum iterations
        for _ in range(len(self.node_data) + 1):
            if len(queue) == 0:
                # no more nodes to visit so we're done
                return collected
            # add the children of this node to be processed
            childs = children.get(queue.pop())
            if childs is not None:
                queue.extend(childs)
                collected.update(childs)
        return collected

    def __hash__(self):
        """
        Actually hash all of the data, but use a "dirty" mechanism
        in functions that modify the data, which MUST
        # all invalidate the hash by setting `self._hash = None`

        This was optimized a bit, and is evaluating on an
        older laptop on a scene with 77 nodes and 76 edges
        10,000 times in 0.7s which seems fast enough.
        """
        # see if there is an available hash value
        # if you are seeing cache bugs this is the thing
        # to try eliminating because it is very likely that
        # someone somewhere is modifying the data without
        # setting `self._hash = None`
        hashed = getattr(self, "_hash", None)
        if hashed is not None:
            return hashed

        hashed = hash_fast(
            (
                "".join(
                    str(hash(k)) + v.get("geometry", "")
                    for k, v in self.edge_data.items()
                )
                + "".join(
                    str(k) + v.get("geometry", "") for k, v in self.node_data.items()
                )
            ).encode("utf-8")
            + b"".join(
                v["matrix"].tobytes() for v in self.edge_data.values() if "matrix" in v
            )
        )
        self._hash = hashed
        return hashed


def kwargs_to_matrix(
    matrix=None, quaternion=None, translation=None, axis=None, angle=None, **kwargs
):
    """
    Take multiple keyword arguments and parse them
    into a homogeneous transformation matrix.

    Returns
    ---------
    matrix : (4, 4) float
      Homogeneous transformation matrix.
    """
    if matrix is not None:
        # a matrix takes immediate precedence over other options
        return np.array(matrix, dtype=np.float64)
    elif quaternion is not None:
        matrix = quaternion_matrix(quaternion)
    elif axis is not None and angle is not None:
        matrix = rotation_matrix(angle, axis)
    else:
        matrix = np.eye(4)

    if translation is not None:
        # translation can be used in conjunction with any
        # of the methods specifying transforms
        matrix[:3, 3] += translation

    return matrix



"""
gltf_extensions.py
------------------

Extension registry for glTF import/export with scope-based handlers.
Each scope has a TypedDict defining the context passed to handlers.
"""

from typing import Any, Callable, OrderedDict, TypedDict

from ...constants import log
from ...typed import Dict, List, Literal, Optional

# Scopes define where in the glTF load/export process handlers run:
#   material            - after parsing material, can override PBR values
#   texture_source      - when resolving texture image index
#   primitive           - after loading primitive, can add face_attributes
#   primitive_preprocess - before accessor reads, can modify accessors in-place
#   primitive_export    - during mesh export, can compress/modify primitive data
Scope = Literal[
    "material", "texture_source", "primitive", "primitive_preprocess", "primitive_export"
]


# ----------------------------------------------------------------------
# TypedDict contexts for each scope
# ----------------------------------------------------------------------
#
# These TypedDicts define the MINIMUM fields passed to handlers for each scope.
# Additional fields may be added in future versions for new functionality.
#
# FOR FORWARD COMPATIBILITY: Handlers should access only the fields they need
# and ignore unknown fields. The context is passed as a plain dict at runtime,
# so handlers can safely use dict.get() for optional access or simply not
# reference fields they don't need.
#
# Example handler pattern:
#
#     def my_handler(context: MaterialContext) -> Optional[Dict]:
#         # Access only what you need - additional fields won't break this
#         data = context["data"]
#         images = context["images"]
#         return {"baseColorFactor": [1, 0, 0, 1]}
#
# ----------------------------------------------------------------------


class MaterialContext(TypedDict):
    """Context for material scope handlers."""

    data: Dict[str, Any]
    parse_textures: Callable[..., Dict[str, Any]]
    images: List


class TextureSourceContext(TypedDict):
    """Context for texture_source scope handlers."""

    data: Dict[str, Any]


class PrimitiveContext(TypedDict):
    """Context for primitive scope handlers (post-load)."""

    data: Dict[str, Any]
    primitive: Dict
    mesh_kwargs: Dict
    accessors: List


class PrimitivePreprocessContext(TypedDict):
    """Context for primitive_preprocess scope handlers (pre-load)."""

    data: Dict[str, Any]
    primitive: Dict
    accessors: List
    views: List


class PrimitiveExportContext(TypedDict):
    """Context for primitive_export scope handlers (during export)."""

    mesh: Any
    name: str
    tree: Dict
    buffer_items: OrderedDict
    primitive: Dict
    include_normals: bool


# Handler type alias - handlers receive a context dict
Handler = Callable[[Any], Any]

# callback to parse material dict and resolve texture references
# signature: (*, data: Dict) -> Dict
ParseTextures = Callable[..., Dict[str, Any]]

# Registry: {scope: {extension_name: handler}}
_handlers: Dict[str, Dict[str, Handler]] = {}


def _deep_merge(target: Dict, source: Dict) -> None:
    """
    Recursively merge source dict into target dict.

    Parameters
    ----------
    target
      Dict to merge into (modified in place)
    source
      Dict to merge from
    """
    for key, value in source.items():
        if isinstance(value, dict) and key in target and isinstance(target[key], dict):
            # Both are dicts - recurse
            _deep_merge(target[key], value)
        else:
            # Overwrite or set new key
            target[key] = value


def register_handler(name: str, scope: Scope) -> Callable[[Handler], Handler]:
    """
    Decorator to register a handler for a glTF extension.

    Parameters
    ----------
    name
      Extension name, e.g. "KHR_materials_pbrSpecularGlossiness".
    scope
      Handler scope, e.g. "material", "texture_source", "primitive".

    Returns
    -------
    decorator
      Function that registers the handler and returns it unchanged.

    Example
    -------
    >>> @register_handler("MY_extension", scope="material")
    ... def my_handler(context: MaterialContext) -> Optional[Dict]:
    ...     data = context["data"]
    ...     images = context["images"]
    ...     return {"baseColorFactor": [1, 0, 0, 1]}
    """
    if scope not in _handlers:
        _handlers[scope] = {}

    def decorator(func: Handler) -> Handler:
        _handlers[scope][name] = func
        return func

    return decorator


def handle_extensions(
    *,
    extensions: Optional[Dict[str, Any]],
    scope: Scope,
    **kwargs,
) -> Any:
    """
    Process extensions dict for a given scope, calling registered handlers.

    Parameters
    ----------
    extensions
      The "extensions" dict from a glTF element, or None.
    scope
      Handler scope to invoke.
    **kwargs
      Scope-specific arguments that will be combined with extension data
      into a typed context dict. Required kwargs by scope:
        - material: parse_textures, images
        - texture_source: (none)
        - primitive: primitive, mesh_kwargs, accessors
        - primitive_preprocess: primitive, accessors, views
        - primitive_export: mesh, name, tree, buffer_items, primitive, include_normals

    Returns
    -------
    results
      Dict of {extension_name: result} for most scopes.
      For scopes ending in "_source", returns first non-None result.
      For "primitive" scope, automatically merges results into mesh_kwargs.
    """
    if not extensions or scope not in _handlers:
        return {} if not scope.endswith("_source") else None

    results = {}
    for ext_name, data in extensions.items():
        if ext_name not in _handlers[scope]:
            continue
        try:
            # Build context dict with data + all kwargs
            context = {"data": data, **kwargs}
            if (result := _handlers[scope][ext_name](context)) is not None:
                results[ext_name] = result
        except Exception as e:
            log.warning(f"failed to process extension {ext_name}: {e}")

    # for _source scopes return first result, otherwise return all results
    if scope.endswith("_source"):
        return next(iter(results.values()), None)

    # for primitive scope, automatically merge results into mesh_kwargs
    if scope == "primitive" and "mesh_kwargs" in kwargs:
        mesh_kwargs = kwargs["mesh_kwargs"]
        for ext_result in results.values():
            if not isinstance(ext_result, dict):
                continue
            # merge extension results, recursively merging nested dicts
            for key, value in ext_result.items():
                if isinstance(value, dict):
                    if key not in mesh_kwargs:
                        mesh_kwargs[key] = {}
                    _deep_merge(mesh_kwargs[key], value)
                else:
                    mesh_kwargs[key] = value

    return results


# ----------------------------------------------------------------------
# Built-in handlers
# ----------------------------------------------------------------------


@register_handler("KHR_materials_pbrSpecularGlossiness", scope="material")
def _specular_glossiness(context: MaterialContext) -> Optional[Dict[str, Any]]:
    """
    Convert specular-glossiness material to PBR metallic-roughness.

    Parameters
    ----------
    context
      MaterialContext with extension data, parse_textures function, and images.

    Returns
    -------
    pbr_dict
      PBR metallic-roughness parameters, or None on failure.
    """
    try:
        from ...visual.gloss import specular_to_pbr

        return specular_to_pbr(**context["parse_textures"](data=context["data"]))
    except Exception:
        log.debug("failed to convert specular-glossiness", exc_info=True)
        return None


@register_handler("EXT_texture_webp", scope="texture_source")
def _texture_webp_source(context: TextureSourceContext) -> Optional[int]:
    """
    Return image source index from EXT_texture_webp.

    Parameters
    ----------
    context
      TextureSourceContext with extension data.

    Returns
    -------
    source_index
      Index into glTF images array, or None if not present.
    """
    return context["data"].get("source")



"""
curvature.py
---------------

Query mesh curvature.
"""

import numpy as np

from . import util
from .util import diagonal_dot

try:
    from scipy.sparse import coo_matrix
except ImportError as E:
    from . import exceptions

    coo_matrix = exceptions.ExceptionWrapper(E)


def face_angles_sparse(mesh):
    """
    A sparse matrix representation of the face angles.

    Returns
    ----------
    sparse : scipy.sparse.coo_matrix
      matrix is float shaped (len(vertices), len(faces))
    """
    matrix = coo_matrix(
        (mesh.face_angles.flatten(), (mesh.faces_sparse.row, mesh.faces_sparse.col)),
        mesh.faces_sparse.shape,
    )
    return matrix


def vertex_defects(mesh):
    """
    Return the vertex defects, or (2*pi) minus the sum of the
    angles of every face that includes that vertex.

    If a vertex is only included by coplanar triangles, this
    will be zero. For convex regions this is positive, and
    concave negative.

    Returns
    --------
    vertex_defect : (len(self.vertices), ) float
                     Vertex defect at the every vertex
    """
    angle_sum = np.array(mesh.face_angles_sparse.sum(axis=1)).flatten()
    defect = (2 * np.pi) - angle_sum
    return defect


def discrete_gaussian_curvature_measure(mesh, points, radius):
    """
    Return the discrete gaussian curvature measure of a sphere
    centered at a point as detailed in 'Restricted Delaunay
    triangulations and normal cycle'- Cohen-Steiner and Morvan.

    This is the sum of the vertex defects at all vertices
    within the radius for each point.

    Parameters
    ----------
    points : (n, 3) float
      Points in space
    radius : float ,
      The sphere radius, which can be zero if vertices
      passed are points.

    Returns
    --------
    gaussian_curvature:  (n,) float
      Discrete gaussian curvature measure.
    """

    points = np.asanyarray(points, dtype=np.float64)
    if not util.is_shape(points, (-1, 3)):
        raise ValueError("points must be (n,3)!")

    nearest = mesh.kdtree.query_ball_point(points, radius)
    gauss_curv = [mesh.vertex_defects[vertices].sum() for vertices in nearest]

    return np.asarray(gauss_curv)


def discrete_mean_curvature_measure(mesh, points, radius):
    """
    Return the discrete mean curvature measure of a sphere
    centered at a point as detailed in 'Restricted Delaunay
    triangulations and normal cycle'- Cohen-Steiner and Morvan.

    This is the sum of the angle at all edges contained in the
    sphere for each point.

    Parameters
    ----------
    points : (n, 3) float
      Points in space
    radius : float
      Sphere radius which should typically be greater than zero

    Returns
    --------
    mean_curvature : (n,) float
      Discrete mean curvature measure.
    """

    points = np.asanyarray(points, dtype=np.float64)
    if not util.is_shape(points, (-1, 3)):
        raise ValueError("points must be (n,3)!")

    # axis aligned bounds
    bounds = np.column_stack((points - radius, points + radius))

    # line segments that intersect axis aligned bounding box
    candidates = [list(mesh.face_adjacency_tree.intersection(b)) for b in bounds]

    mean_curv = np.zeros(len(points))
    for i, (x, x_candidates) in enumerate(zip(points, candidates)):
        endpoints = mesh.vertices[mesh.face_adjacency_edges[x_candidates]]
        lengths = line_ball_intersection(
            endpoints[:, 0], endpoints[:, 1], center=x, radius=radius
        )
        angles = mesh.face_adjacency_angles[x_candidates]
        signs = np.where(mesh.face_adjacency_convex[x_candidates], 1, -1)
        mean_curv[i] = (lengths * angles * signs).sum() / 2

    return mean_curv


def line_ball_intersection(start_points, end_points, center, radius):
    """
    Compute the length of the intersection of a line segment with a ball.

    Parameters
    ----------
    start_points : (n,3) float, list of points in space
    end_points   : (n,3) float, list of points in space
    center       : (3,) float, the sphere center
    radius       : float, the sphere radius

    Returns
    --------
    lengths: (n,) float, the lengths.

    """

    # We solve for the intersection of |x-c|**2 = r**2 and
    # x = o + dL. This yields
    # d = (-l.(o-c) +- sqrt[ l.(o-c)**2 - l.l((o-c).(o-c) - r^**2) ]) / l.l
    L = end_points - start_points
    oc = start_points - center  # o-c
    r = radius
    ldotl = diagonal_dot(L, L)
    ldotoc = diagonal_dot(L, oc)
    ocdotoc = diagonal_dot(oc, oc)
    discrims = ldotoc**2 - ldotl * (ocdotoc - r**2)

    # If discriminant is non-positive, then we have zero length
    lengths = np.zeros(len(start_points))
    # Otherwise we solve for the solns with d2 > d1.
    m = discrims > 0  # mask
    d1 = (-ldotoc[m] - np.sqrt(discrims[m])) / ldotl[m]
    d2 = (-ldotoc[m] + np.sqrt(discrims[m])) / ldotl[m]

    # Line segment means we have 0 <= d <= 1
    d1 = np.clip(d1, 0, 1)
    d2 = np.clip(d2, 0, 1)

    # Length is |o + d2 l - o + d1 l|  = (d2 - d1) |l|
    lengths[m] = (d2 - d1) * np.sqrt(ldotl[m])

    return lengths


def sphere_ball_intersection(R, r):
    """
    Compute the surface area of the intersection of sphere of radius R centered
    at (0, 0, 0) with a ball of radius r centered at (R, 0, 0).

    Parameters
    ----------
    R : float, sphere radius
    r : float, ball radius

    Returns
    --------
    area: float, the surface are.
    """
    x = (2 * R**2 - r**2) / (2 * R)  # x coord of plane
    if x >= -R:
        return 2 * np.pi * R * (R - x)
    if x < -R:
        return 4 * np.pi * R**2



def test_line_closed():
    import numpy as np

    from trimesh.path.entities import Line

    e = Line(points=[0, 1, 2], closed=False)
    assert np.allclose(e.points, [0, 1, 2])

    e = Line(points=[0, 1, 2], closed=True)
    assert np.allclose(e.points, [0, 1, 2, 0])

    e = Line(points=[0, 1, 2, 0], closed=True)
    assert np.allclose(e.points, [0, 1, 2, 0])

    # should it really drop the last point... that seems weird
    e = Line(points=[0, 1, 2, 0], closed=False)
    assert np.allclose(e.points, [0, 1, 2, 0])


if __name__ == "__main__":
    import trimesh

    trimesh.util.attach_to_log()
    test_line_closed()



try:
    from . import generic as g
except BaseException:
    import generic as g


def test_svg():
    for d in g.get_2D():
        if g.np.isclose(d.area, 0.0):
            continue
        # export and reload the exported SVG
        loaded = g.trimesh.load(
            g.trimesh.util.wrap_as_stream(d.export(file_type="svg")), file_type="svg"
        )

        # we only have line and arc primitives as SVG
        # export and import
        if all(i.__class__.__name__ in ["Line", "Arc"] for i in d.entities):
            # perimeter should stay the same-ish
            # on export/import
            assert g.np.isclose(d.length, loaded.length, rtol=0.01)

            assert len(d.entities) == len(loaded.entities)

            path_str = g.trimesh.path.exchange.svg_io.export_svg(d, return_path=True)
            assert isinstance(path_str, str)
            assert len(path_str) > 0


def test_layer():
    from shapely.geometry import Point

    # create two disjoint circles and apply layers
    a = g.trimesh.load_path(Point([0, 0]).buffer(1))
    a.apply_layer("ACIRCLE")

    b = g.trimesh.load_path(Point([2, 0]).buffer(1))
    b.apply_layer("BCIRCLE")

    assert id(a.entities[0]._metadata) != id(b.entities[0]._metadata)

    # combine two circles
    c = a + b

    # should be combined
    assert g.np.isclose(c.area, a.area + b.area)

    # export C with just layer of A
    aX = g.trimesh.load(
        g.io_wrap(c.export(file_type="svg", only_layers=["ACIRCLE"])), file_type="svg"
    )

    # export C with all layers
    cX = g.trimesh.load(
        g.io_wrap(c.export(file_type="svg", only_layers=None)), file_type="svg"
    )

    assert len(cX.entities) == len(c.entities)
    # should have skipped the layers
    assert len(aX.entities) == 1

    # make
    aR = g.trimesh.load(
        g.io_wrap(c.export(file_type="dxf", only_layers=["ACIRCLE"])), file_type="dxf"
    )

    assert g.np.isclose(aR.area, a.area)


def test_trans():
    from trimesh.path.exchange.svg_io import transform_to_matrices as tf

    # empty strings shouldn't have matrix
    assert len(tf("")) == 0

    # check translate with different whitespace
    a = tf("translate(1.1, 2.2      )")
    assert len(a) == 1
    assert g.np.allclose(a[0], [[1, 0, 1.1], [0, 1, 2.2], [0, 0, 1]])
    a = tf(" translate(1.1    1.2   )       ")
    assert len(a) == 1
    assert g.np.allclose(a[0], [[1, 0, 1.1], [0, 1, 1.2], [0, 0, 1]])

    a = tf(
        " translate(1.1    1.2   )       "
        + "matrix (  {} {} {} {} {} {})".format(*g.np.arange(6))
    )
    assert len(a) == 2
    # check the translate
    assert g.np.allclose(a[0], [[1, 0, 1.1], [0, 1, 1.2], [0, 0, 1]])
    # check the matrix string
    assert g.np.allclose(a[1], [[0, 2, 4], [1, 3, 5], [0, 0, 1]])


def test_roundtrip():
    """
    Check to make sure a roundtrip from both a Scene and a
    Path2D results in the same file on both sides
    """
    for fn in ["2D/250_cycloidal.DXF", "2D/tray-easy1.dxf"]:
        p = g.get_mesh(fn)
        assert isinstance(p, g.trimesh.path.Path2D)
        # load the exported SVG
        r = g.trimesh.load(
            g.trimesh.util.wrap_as_stream(p.export(file_type="svg")), file_type="svg"
        )
        assert isinstance(r, g.trimesh.path.Path2D)
        assert g.np.isclose(r.length, p.length)
        assert g.np.isclose(r.area, p.area)

        assert set(r.metadata.keys()) == set(p.metadata.keys())

        s = g.trimesh.scene.split_scene(p)

        as_svg = s.export(file_type="svg")
        assert isinstance(s, g.trimesh.Scene)
        r = g.trimesh.load(g.trimesh.util.wrap_as_stream(as_svg), file_type="svg")
        assert isinstance(r, g.trimesh.Scene)
        assert s.metadata == r.metadata

        # make sure every geometry name matches exactly
        assert set(s.geometry.keys()) == set(r.geometry.keys())

        # check to see if every geometry has the same metadata
        for geom in s.geometry.keys():
            a, b = s.geometry[geom], r.geometry[geom]
            assert a.metadata == b.metadata
            assert g.np.isclose(a.length, b.length)
            assert g.np.isclose(a.area, b.area)
            assert a.body_count == b.body_count

        # assert r.metadata["file_path"].endswith(fn[3:])


def test_shapes():
    # the "basic shapes" example from the mozilla SVG docs
    p = g.trimesh.load_path(
        g.trimesh.util.wrap_as_stream("""<?xml version="1.0" standalone="no"?>
<svg width="200" height="250" version="1.1" xmlns="http://www.w3.org/2000/svg">

  <rect x="10" y="10" width="30" height="30" stroke="black" fill="transparent" stroke-width="5"/>
  <rect x="60" y="10" rx="10" ry="10" width="30" height="30" stroke="black" fill="transparent" stroke-width="5"/>

  <circle cx="25" cy="75" r="20" stroke="red" fill="transparent" stroke-width="5"/>
  <ellipse cx="75" cy="75" rx="20" ry="5" stroke="red" fill="transparent" stroke-width="5"/>

  <line x1="10" x2="50" y1="110" y2="150" stroke="orange" stroke-width="5"/>
  <polyline points="60 110 65 120 70 115 75 130 80 125 85 140 90 135 95 150 100 145"
      stroke="orange" fill="transparent" stroke-width="5"/>

  <polygon points="50 160 55 180 70 180 60 190 65 205 50 195 35 205 40 190 30 180 45 180"
      stroke="green" fill="transparent" stroke-width="5"/>

  <path d="M20,230 Q40,205 50,230 T90,230" fill="none" stroke="blue" stroke-width="5"/>
</svg>
"""),
        file_type="svg",
    )

    # should have gotten the closed circle
    assert any(e.closed for e in p.entities if isinstance(e, g.trimesh.path.entities.Arc))

    assert len(p.entities) >= 8


def test_rect_bounds():
    """
    Check to make sure we're handling rectangle scale correctly
    """

    def get_bounds(path) -> set:
        # get the bounds of every entity as a set of flat integer tuples
        return {
            tuple(e.bounds(path.vertices).ravel().round().astype(int).tolist())
            for e in path.entities
        }

    # pairs of svg file with shapes and transforms
    # and an evaluated-by-inkscape path string with all transforms fully baked
    truths = g.data["svg_transform_truth"]
    assert len(truths) >= 7

    for pair in truths:
        # shapes with scale, translation, etc
        ours = g.trimesh.load_path(
            g.trimesh.util.wrap_as_stream(pair["svg"]),
            file_type="svg",
        )

        # converted to a path string using inkscape, i.e. "truth"
        truth = g.trimesh.path.Path2D(
            **g.trimesh.path.exchange.svg_io.svg_to_path(path_string=pair["truth"])
        )
        # make sure we don't accept empty results
        assert len(ours.entities) > 0

        # the set of entity AABB should match inkscape's output exactly
        bounds_ours = get_bounds(ours)
        bounds_truth = get_bounds(truth)
        if bounds_ours != bounds_truth:
            raise ValueError(f"{bounds_ours} != {bounds_truth}\nfor:\n{pair['truth']}")


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    # test_roundtrip()
    test_rect_bounds()



def test_exception_wrapper():
    # check our exception wrapping behavior
    from trimesh.exceptions import ExceptionWrapper

    try:
        # create a wrapper around a ValueError
        raise ValueError("This should be wrapped!")
    except BaseException as E:
        wrapper = ExceptionWrapper(E)

    # `isinstance` should NOT raise the error that's been wrapped
    assert not isinstance(wrapper, dict)

    # but we should be able to tell that it is wrapping something
    assert isinstance(wrapper, ExceptionWrapper)

    try:
        # check the __getattribute__ path
        wrapper.hi()

        # we wrapped something incorrectly if we got here!
        raise NameError("__getattribute__ should have raised a ValueError!")
    except ValueError:
        # should have re-raised ValueError
        pass

    try:
        # check the __call__ path
        wrapper()

        # we wrapped something incorrectly if we got here!
        raise NameError("__call__ should have raised a ValueError!")

    except ValueError:
        # this is correct
        pass


if __name__ == "__main__":
    test_exception_wrapper()



try:
    from . import generic as g
except BaseException:
    import generic as g


class ArcTests(g.unittest.TestCase):
    def test_center(self):
        from trimesh.path.arc import arc_center

        test_points = [[[0, 0], [1.0, 1], [2, 0]]]
        test_results = [[[1, 0], 1.0]]
        points = test_points[0]
        res_center, res_radius = test_results[0]
        center_info = arc_center(points)
        C, R, _N, _angle = (
            center_info["center"],
            center_info["radius"],
            center_info["normal"],
            center_info["span"],
        )

        assert abs(R - res_radius) < g.tol_path.zero
        assert g.np.linalg.norm(C - res_center) < g.tol_path.zero
        # large magnitude arc failed some coplanar tests
        c = g.trimesh.path.arc.arc_center(
            [
                [30156.18, 1673.64, -2914.56],
                [30152.91, 1780.09, -2885.51],
                [30148.3, 1875.81, -2857.79],
            ]
        )
        assert len(c.center) == 3

    def test_center_random(self):
        from trimesh.path.arc import arc_center

        # Test that arc centers work on well formed random points in 2D and 3D
        min_angle = g.np.radians(2)
        count = 1000

        center_3D = (g.random((count, 3)) - 0.5) * 50
        center_2D = center_3D[:, 0:2]
        radii = g.np.clip(g.random(count) * 100, min_angle, g.np.inf)

        angles = g.random((count, 2)) * (g.np.pi - min_angle) + min_angle
        angles = g.np.column_stack((g.np.zeros(count), g.np.cumsum(angles, axis=1)))

        points_2D = g.np.column_stack(
            (
                g.np.cos(angles[:, 0]),
                g.np.sin(angles[:, 0]),
                g.np.cos(angles[:, 1]),
                g.np.sin(angles[:, 1]),
                g.np.cos(angles[:, 2]),
                g.np.sin(angles[:, 2]),
            )
        ).reshape((-1, 6))
        points_2D *= radii.reshape((-1, 1))
        points_2D += g.np.tile(center_2D, (1, 3))
        points_2D = points_2D.reshape((-1, 3, 2))
        points_3D = g.np.column_stack(
            (
                points_2D.reshape((-1, 2)),
                g.np.tile(center_3D[:, 2].reshape((-1, 1)), (1, 3)).reshape(-1),
            )
        ).reshape((-1, 3, 3))
        for center, radius, three in zip(center_2D, radii, points_2D):
            info = arc_center(three)

            assert g.np.allclose(center, info["center"])
            assert g.np.allclose(radius, info["radius"])

        for center, radius, three in zip(center_3D, radii, points_3D):
            transform = g.trimesh.transformations.random_rotation_matrix()
            center = g.trimesh.transformations.transform_points([center], transform)[0]
            three = g.trimesh.transformations.transform_points(three, transform)

            info = arc_center(three)

            assert g.np.allclose(center, info["center"])
            assert g.np.allclose(radius, info["radius"])

    def test_multiroot(self):
        """
        Test a Path2D object containing polygons nested in
        the interiors of other polygons.
        """
        inner = g.trimesh.creation.annulus(r_min=0.5, r_max=0.6, height=1.0)
        outer = g.trimesh.creation.annulus(r_min=0.9, r_max=1.0, height=1.0)
        m = inner + outer

        s = m.section(plane_normal=[0, 0, 1], plane_origin=[0, 0, 0])
        p = s.to_planar()[0]

        assert len(p.polygons_closed) == 4
        assert len(p.polygons_full) == 2
        assert len(p.root) == 2
        g.check_path2D(p)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
repair.py
-------------

Fill holes and fix winding and normals of meshes.
"""

import numpy as np

from . import graph, triangles
from .constants import log
from .geometry import faces_to_edges, triangulate_quads
from .graph import connected_components
from .grouping import group_rows, hashable_rows

try:
    import networkx as nx
except BaseException as E:
    # create a dummy module which will raise the ImportError
    # or other exception only when someone tries to use networkx
    from .exceptions import ExceptionWrapper

    nx = ExceptionWrapper(E)

try:
    from .path.exchange.misc import faces_to_path
except BaseException as E:
    from .exceptions import ExceptionWrapper

    faces_to_path = ExceptionWrapper(E)


def fix_winding(mesh):
    """
    Traverse and change mesh faces in-place to make sure
    winding is correct with edges on adjacent faces in
    opposite directions.

    Parameters
    -------------
    mesh : Trimesh
      Source geometry to alter in-place.
    """
    # anything we would fix is already done
    if mesh.is_winding_consistent:
        return

    graph_all = nx.from_edgelist(mesh.face_adjacency)
    flipped = 0

    faces = mesh.faces.view(np.ndarray).copy()

    # we are going to traverse the graph using BFS
    # start a traversal for every connected component
    for components in nx.connected_components(graph_all):
        # get a subgraph for this component
        g = graph_all.subgraph(components)
        # get the first node in the graph in a way that works on nx's
        # new API and their old API
        start = next(iter(g.nodes()))

        # we traverse every pair of faces in the graph
        # we modify mesh.faces and mesh.face_normals in place
        for face_pair in nx.bfs_edges(g, start):
            # for each pair of faces, we convert them into edges,
            # find the edge that both faces share and then see if edges
            # are reversed in order as you would expect
            # (2, ) int
            face_pair = np.ravel(face_pair)
            # (2, 3) int
            pair = faces[face_pair]
            # (6, 2) int
            edges = faces_to_edges(pair)
            overlap = group_rows(np.sort(edges, axis=1), require_count=2)
            if len(overlap) == 0:
                # only happens on non-watertight meshes
                continue
            edge_pair = edges[overlap[0]]
            if edge_pair[0][0] == edge_pair[1][0]:
                # if the edges aren't reversed, invert the order of one face
                flipped += 1
                faces[face_pair[1]] = faces[face_pair[1]][::-1]

    if flipped > 0:
        mesh.faces = faces

    log.debug("flipped %d/%d edges", flipped, len(mesh.faces) * 3)


def fix_inversion(mesh, multibody: bool = False):
    """
    Check to see if a mesh has normals pointing "out."

    Parameters
    -------------
    mesh : trimesh.Trimesh
      Mesh to fix in-place.
    multibody : bool
      If True will try to fix normals on every body
    """

    # we are not considering multiple bodies so we only need to
    # do anything if the mesh is watertight and the volume is negative
    if not multibody:
        if mesh.is_watertight and mesh.volume < 0.0:
            # this will make things worse for non-watertight meshes
            mesh.invert()
        return

    # get groups of connected faces
    groups = connected_components(mesh.face_adjacency)

    # mask of faces to flip
    flip = np.zeros(len(mesh.faces), dtype=bool)

    # save these to avoid thrashing cache in a loop
    tri = mesh.triangles
    cross = mesh.triangles_cross
    # get the edges indexable as faces
    face_edges = mesh.edges.reshape((-1, 6))

    # loop through indexes of `mesh.faces`
    for face_index in groups:
        # we need watertight bodies to do anything so immediately
        # skip anything smaller than 4 faces (tetrahedron)
        if len(face_index) < 4:
            continue
        # check the watertightness and winding of this group
        is_tight, is_wound = graph.is_watertight(face_edges[face_index].reshape((-1, 2)))

        # if we're not completely correct skip it as flipping faces will make it worse
        if not (is_tight and is_wound):
            continue

        # check the volume for just this body
        volume = triangles.mass_properties(
            tri[face_index], crosses=cross[face_index], skip_inertia=True
        ).volume
        # if that volume is negative it is either
        # inverted or just total garbage
        if volume < 0.0:
            flip[face_index] = True

    # one or more faces needs flipping
    if flip.any():
        if "face_normals" in mesh._cache:
            # flip normals of necessary faces
            normals = mesh.face_normals.copy()
            normals[flip] *= -1.0
        else:
            normals = None
        # flip faces
        mesh.faces[flip] = np.fliplr(mesh.faces[flip])
        # assign corrected normals
        if normals is not None:
            mesh.face_normals = normals


def fix_normals(mesh, multibody=False):
    """
    Fix the winding and direction of a mesh face and
    face normals in-place.

    Really only meaningful on watertight meshes but will orient all
    faces and winding in a uniform way for non-watertight face
    patches as well.

    Parameters
    -------------
    mesh : trimesh.Trimesh
      Mesh to fix normals on
    multibody : bool
      if True try to correct normals direction
      on every body rather than just one

    Notes
    --------------
    mesh.faces : will flip columns on inverted faces
    """
    # traverse face adjacency to correct winding
    fix_winding(mesh)
    # check to see if a mesh is inverted
    fix_inversion(mesh, multibody=multibody)


def broken_faces(mesh, color=None):
    """
    Return the index of faces in the mesh which break the
    watertight status of the mesh.

    Parameters
    --------------
    mesh : trimesh.Trimesh
      Mesh to check broken faces on
    color: (4,) uint8 or None
      Will set broken faces to this color if not None

    Returns
    ---------------
    broken : (n, ) int
      Indexes of mesh.faces
    """
    adjacency = nx.from_edgelist(mesh.face_adjacency)
    broken = [k for k, v in dict(adjacency.degree()).items() if v != 3]
    broken = np.array(broken)
    if color is not None and broken.size != 0:
        # if someone passed a broken color
        color = np.array(color)
        if not (color.shape == (4,) or color.shape == (3,)):
            color = [255, 0, 0, 255]
        mesh.visual.face_colors[broken] = color
    return broken


def fill_holes(mesh, use_fan: bool = False):
    """
    Fill boundary holes in-place using fans, which may result
    in bad answers if the holes are non convex!

    Face colors and attributes will be padded with default values
    so shapes match.

    Parameters
    ----------
    mesh : trimesh.Trimesh
        Mesh will be repaired in-place.
    use_fan
      If passed, holes larger than quads will be triangulated
      using fans which are only valid for non-convex holes.
    """
    if len(mesh.faces) < 3:
        return False
    if mesh.is_watertight:
        return True

    # any edge that occurs once is on the boundary
    boundary_groups = group_rows(mesh.edges_sorted, require_count=1)
    # either watertight or broken in a weird way
    if len(boundary_groups) < 3:
        return False

    # ordered edges on the boundary
    boundary = mesh.edges[boundary_groups]

    # use networkx to find cycles of boundary edges
    holes = nx.cycle_basis(nx.from_edgelist(boundary))

    # this handles mixed tris, quads, and arbitrary polygons
    new_faces = triangulate_quads(holes, use_fan=use_fan)
    if len(new_faces) == 0:
        return False

    # now we need to fix the winding for every new face
    new_edges = faces_to_edges(new_faces)

    # the hashable rows for the mesh boundary edges
    # and the boundary of the original hole in the mesh
    # a well-constructed mesh has edges that are REVERSED
    # so we're going to try a cheap indexing operation
    hashable_new = hashable_rows(new_edges)
    hashable_old = hashable_rows(boundary)

    # the magic trick: if any ordered edge of the NEW face is contained
    # in the OLD boundary edges, it means the new face has edges that
    # are in the same order as the boundary, and that new face
    # needs to be reversed and since new_edges has exactly three
    # edges per triangle we can compact indexes back to triangles
    needs_reverse = np.isin(hashable_new, hashable_old).reshape((-1, 3)).any(axis=1)

    # now we have some shiny new faces that might even be wound correctly!
    new_faces[needs_reverse] = np.fliplr(new_faces[needs_reverse])

    # do the bookkeeping to preserve face normals and pad attributes
    mesh.extend_faces(new_faces)

    return mesh.is_watertight


def stitch(mesh, faces=None, insert_vertices=False):
    """
    Create a fan stitch over the boundary of the specified
    faces. If the boundary is non-convex a triangle fan
    is going to be extremely wonky.

    Parameters
    -----------
    mesh : trimesh.Trimesh
      Mesh to create fan stitch on.
    faces : (n,) int
      Face indexes to stitch with triangle fans.
    insert_vertices : bool
      Allow stitching to insert new vertices?

    Returns
    ----------
    fan : (m, 3) int
      New triangles referencing mesh.vertices.
    vertices : (p, 3) float
      Inserted vertices (only returned `if insert_vertices`)
    """
    if faces is None:
        faces = np.arange(len(mesh.faces))

    # get a sequence of vertex indices representing the
    # boundary of the specified faces
    # will be referencing the same indexes of `mesh.vertices`
    points = [
        e.points
        for e in faces_to_path(mesh, faces)["entities"]
        if len(e.points) > 3 and e.points[0] == e.points[-1]
    ]

    # get properties to avoid querying in loop
    vertices = mesh.vertices
    normals = mesh.face_normals

    # find which faces are associated with an edge
    edges_face = mesh.edges_face
    tree_edge = mesh.edges_sorted_tree

    if insert_vertices:
        # create one new vertex per curve at the centroid
        centroids = np.array([vertices[p].mean(axis=0) for p in points])
        # save the original length of the vertices
        count = len(vertices)
        # for the normal check stack our local vertices
        vertices = np.vstack((vertices, centroids))
        # create a triangle between our new centroid vertex
        # and each one of the boundary curves
        fan = [
            np.column_stack((np.ones(len(p) - 1, dtype=int) * (count + i), p[:-1], p[1:]))
            for i, p in enumerate(points)
        ]
    else:
        # since we're not allowed to insert new vertices
        # create a triangle fan for each boundary curve
        fan = [
            np.column_stack((np.ones(len(p) - 3, dtype=int) * p[0], p[1:-2], p[2:-1]))
            for p in points
        ]

    # now we do a normal check against an adjacent face
    # to see if each region needs to be flipped
    for i, t in zip(range(len(fan)), fan):
        # get the edges from the original mesh
        # for the first `n` new triangles
        e = t[:10, 1:].copy()
        e.sort(axis=1)

        # find which indexes of `mesh.edges` these
        # new edges correspond with by finding edges
        # that exactly correspond with the tree
        query = tree_edge.query_ball_point(e, r=1e-10)
        if len(query) == 0:
            continue
        # stack all the indices that exist
        edge_index = np.concatenate(query)

        # get the normals from the original mesh
        original = normals[edges_face[edge_index]]

        # calculate the normals for a few new faces
        check, valid = triangles.normals(vertices[t[:3]])
        if not valid.any():
            continue
        # take the first valid normal from our new faces
        check = check[0]

        # if our new faces are reversed from the original
        # Adjacent face flip them along their axis
        sign = np.dot(original, check)
        if sign.mean() < 0:
            fan[i] = np.fliplr(t)

    fan = np.vstack(fan)

    if insert_vertices:
        return fan, centroids
    return fan



"""
Test for PBR material concatenation bug fix.

Tests that concatenating meshes with PBR materials preserves
metallicFactor and roughnessFactor values and creates correct
UV mappings for textures.
"""

try:
    from . import generic as g
except BaseException:
    import generic as g


class PBRConcatenateTest(g.unittest.TestCase):
    def test_pbr_factors_preserved(self):
        """Test that metallicFactor and roughnessFactor are preserved when
        concatenating meshes with PBR materials.

        This tests the fix for the bug where these factors became None after
        concatenation, causing meshes to appear shiny/metallic.
        """
        # Create two boxes with PBR materials
        box1 = g.trimesh.creation.box(extents=[1, 1, 1])
        mat1 = g.trimesh.visual.material.PBRMaterial(
            baseColorFactor=[255, 51, 51, 255],  # Red
            metallicFactor=0.0,
            roughnessFactor=1.0,
        )
        box1.visual = g.trimesh.visual.TextureVisuals(material=mat1)

        box2 = g.trimesh.creation.box(extents=[1, 1, 1])
        mat2 = g.trimesh.visual.material.PBRMaterial(
            baseColorFactor=[51, 255, 51, 255],  # Green
            metallicFactor=0.0,
            roughnessFactor=1.0,
        )
        box2.visual = g.trimesh.visual.TextureVisuals(material=mat2)
        box2.apply_translation([1.5, 0, 0])

        # Concatenate the meshes
        merged = g.trimesh.util.concatenate([box1, box2])

        # Factor are `None` because textures are created for metallicFactor and
        # roughnessFactor
        assert merged.visual.material.metallicFactor is None
        assert merged.visual.material.roughnessFactor is None

        # Check that textures were created
        assert merged.visual.material.baseColorTexture is not None
        assert merged.visual.material.metallicRoughnessTexture is not None

    def test_pbr_uv_mapping(self):
        """Test that UV coordinates correctly map to material values in the
        packed textures after concatenation.
        """
        # Create two boxes with different colors but same PBR values
        box1 = g.trimesh.creation.box(extents=[0.5, 0.5, 0.5])
        mat1 = g.trimesh.visual.material.PBRMaterial(
            baseColorFactor=[255, 0, 0, 255],
            metallicFactor=0.0,
            roughnessFactor=1.0,
        )
        box1.visual = g.trimesh.visual.TextureVisuals(material=mat1)

        box2 = g.trimesh.creation.box(extents=[0.5, 0.5, 0.5])
        mat2 = g.trimesh.visual.material.PBRMaterial(
            baseColorFactor=[0, 255, 0, 255],
            metallicFactor=0.0,
            roughnessFactor=1.0,
        )
        box2.visual = g.trimesh.visual.TextureVisuals(material=mat2)
        box2.apply_translation([1, 0, 0])

        # Concatenate
        merged = g.trimesh.util.concatenate([box1, box2])

        # Get the metallicRoughness texture
        tex = g.np.array(merged.visual.material.metallicRoughnessTexture)

        # Check that all UV coordinates map to the correct material values
        # In glTF: Blue channel = metallic, Green channel = roughness
        for u, v in merged.visual.uv:
            # Convert UV to pixel coordinates
            px = round(u * (tex.shape[1] - 1))
            py = round((1 - v) * (tex.shape[0] - 1))

            if 0 <= px < tex.shape[1] and 0 <= py < tex.shape[0]:
                roughness = tex[py, px, 1]  # Green channel
                metallic = tex[py, px, 2]  # Blue channel

                # All vertices should map to roughness=255 (1.0) and metallic=0 (0.0)
                assert roughness == 255, (
                    f"UV ({u:.3f}, {v:.3f}) maps to wrong roughness: {roughness}"
                )
                assert metallic == 0, (
                    f"UV ({u:.3f}, {v:.3f}) maps to wrong metallic: {metallic}"
                )

    def test_pbr_identical_materials(self):
        """Test that concatenating meshes with identical PBR materials
        preserves the scalar values without creating textures."""
        # Create two boxes with identical materials
        box1 = g.trimesh.creation.box(extents=[0.5, 0.5, 0.5])
        mat1 = g.trimesh.visual.material.PBRMaterial(
            baseColorFactor=[128, 128, 128, 255],
            metallicFactor=0.5,
            roughnessFactor=0.7,
        )
        box1.visual = g.trimesh.visual.TextureVisuals(material=mat1)

        box2 = g.trimesh.creation.box(extents=[0.5, 0.5, 0.5])
        mat2 = g.trimesh.visual.material.PBRMaterial(
            baseColorFactor=[128, 128, 128, 255],  # Same color
            metallicFactor=0.5,  # Same metallic
            roughnessFactor=0.7,  # Same roughness
        )
        box2.visual = g.trimesh.visual.TextureVisuals(material=mat2)
        box2.apply_translation([1, 0, 0])

        # Concatenate
        merged = g.trimesh.util.concatenate([box1, box2])

        # When materials are identical, scalar values should be preserved
        assert merged.visual.material.metallicFactor == 0.5
        assert merged.visual.material.roughnessFactor == 0.7

        # No metallicRoughness texture should be created for identical materials
        assert merged.visual.material.metallicRoughnessTexture is None


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


def test_3MF():
    # an assembly with instancing
    s = g.get_mesh("counterXP.3MF")

    # should be 2 unique meshes
    assert len(s.geometry) == 2
    # should be 6 instances around the scene
    assert len(s.graph.nodes_geometry) == 6
    assert all(m.is_volume for m in s.geometry.values())

    # a single body 3MF assembly
    s = g.get_mesh("featuretype.3MF")
    # should be 2 unique meshes
    assert len(s.geometry) == 1
    # should be 6 instances around the scene
    assert len(s.graph.nodes_geometry) == 1


def test_units():
    # test our unit conversion function
    converter = g.trimesh.units.unit_conversion
    # these are the units listed in the 3MF spec as valid
    units = ["micron", "millimeter", "centimeter", "inch", "foot", "meter"]
    # check conversion factor for all valid 3MF units
    assert all(converter(u, "inches") > 1e-12 for u in units)


def test_more_units():
    # make sure that units survive down to the mesh
    s = g.trimesh.load_scene(g.os.path.join(g.dir_models, "featuretype.3MF"))
    m = next(iter(s.geometry.values()))
    assert m.units is not None, m.metadata


def test_kwargs():
    # check if kwargs are properly passed to geometries
    s = g.get_mesh("P_XPM_0331_01.3mf")
    assert all(len(v.vertices) == 4 for v in s.geometry.values())

    s = g.get_mesh("P_XPM_0331_01.3mf", process=False)
    assert all(len(v.vertices) == 5 for v in s.geometry.values())


def test_names():
    # check if two different objects with the same name are correctly
    # processed
    s = g.get_mesh("cube_and_sphere_same_name.3mf")
    assert len(s.geometry) == 2


def test_roundtrip():
    if g.sys.version_info < (3, 6):
        g.log.warning("relies on > Python 3.5")
        return

    # test a scene round-tripped through the
    # 3MF exporter and importer

    for name, count in [("cycloidal.3DXML", 13), ("pyramids.3mf", 6)]:
        s = g.get_mesh(name)
        assert len(s.geometry) == count

        # export and reload
        r = g.trimesh.load(
            file_obj=g.trimesh.util.wrap_as_stream(s.export(file_type="3mf")),
            file_type="3mf",
        )

        assert set(s.geometry.keys()) == set(r.geometry.keys()), (
            s.geometry.keys(),
            r.geometry.keys(),
        )
        assert g.np.allclose(s.bounds, r.bounds)
        assert g.np.isclose(s.area, r.area, rtol=1e-3)

        if "3mf" in name.lower():
            # check partnumbers are preserved in the input 3mf files
            s_partnumbers = {e[1] for e in s.graph.to_edgelist()}
            r_partnumbers = {e[1] for e in r.graph.to_edgelist()}
            assert s_partnumbers == r_partnumbers, (s_partnumbers, r_partnumbers)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
nsphere.py
--------------

Functions for fitting and minimizing nspheres:
circles, spheres, hyperspheres, etc.
"""

import numpy as np

from . import convex, util
from .constants import log, tol

try:
    # scipy is a soft dependency
    from scipy import spatial
    from scipy.optimize import leastsq
except BaseException as E:
    # raise the exception when someone tries to use it
    from . import exceptions

    leastsq = exceptions.ExceptionWrapper(E)
    spatial = exceptions.ExceptionWrapper(E)


def minimum_nsphere(obj):
    """
    Compute the minimum n- sphere for a mesh or a set of points.

    Uses the fact that the minimum n- sphere will be centered at one of
    the vertices of the furthest site voronoi diagram, which is n*log(n)
    but should be pretty fast due to using the scipy/qhull implementations
    of convex hulls and voronoi diagrams.

    Parameters
    ----------
    obj : (n, d) float or trimesh.Trimesh
      Points or mesh to find minimum bounding nsphere

    Returns
    ----------
    center : (d,) float
      Center of fitted n- sphere
    radius : float
      Radius of fitted n-sphere
    """

    # reduce the input points or mesh to the vertices of the convex hull
    # since we are computing the furthest site voronoi diagram this reduces
    # the input complexity substantially and returns the same value
    points = convex.hull_points(obj)

    # we are scaling the mesh to a unit cube
    # this used to pass qhull_options 'QbB' to Voronoi however this had a bug somewhere
    # to avoid this we scale to a unit cube ourselves inside this function
    points_origin = points.min(axis=0)
    points_scale = np.ptp(points, axis=0).min()
    points = (points - points_origin) / points_scale

    # if all of the points are on an n-sphere already the voronoi
    # method will fail so we check a least squares fit before
    # bothering to compute the voronoi diagram
    fit_C, fit_R, fit_E = fit_nsphere(points)
    # return fit radius and center to global scale
    fit_R = (((points - fit_C) ** 2).sum(axis=1).max() ** 0.5) * points_scale
    fit_C = (fit_C * points_scale) + points_origin

    if fit_E < 1e-6:
        # points were on an n-sphere so just return fit
        return fit_C, fit_R

    # calculate a furthest site voronoi diagram
    # this will fail if the points are ALL on the surface of
    # the n-sphere but hopefully the least squares check caught those cases
    # , qhull_options='QbB Pp')
    voronoi = spatial.Voronoi(points, furthest_site=True)

    # find the maximum radius^2 point for each of the voronoi vertices
    # this is worst case quite expensive but we have taken
    # convex hull to reduce n for this operation
    # we are doing comparisons on the radius squared then rooting once
    try:
        # cdist is massivly faster than looping or tiling methods
        # although it does create an extremely large intermediate array
        # first, get an order of magnitude memory size estimate
        # a float64 would be 8 bytes per entry plus overhead
        memory_estimate = len(voronoi.vertices) * len(points) * 9
        # if this `cdist` would require more than 4gb of memory fall
        # back to the more memory-efficent but much slower loop
        if memory_estimate > 4e9:
            raise MemoryError
        radii_2 = spatial.distance.cdist(
            voronoi.vertices, points, metric="sqeuclidean"
        ).max(axis=1)
    except MemoryError:
        # log the MemoryError
        log.warning("MemoryError: falling back to slower check!")
        # fall back to a potentially very slow list comprehension
        radii_2 = np.array(
            [((points - v) ** 2).sum(axis=1).max() for v in voronoi.vertices]
        )

    # we want the smallest sphere so take the min of the radii
    radii_idx = radii_2.argmin()

    # return voronoi radius and center to global scale
    radius_v = np.sqrt(radii_2[radii_idx]) * points_scale
    center_v = (voronoi.vertices[radii_idx] * points_scale) + points_origin

    if radius_v > fit_R:
        return fit_C, fit_R

    return center_v, radius_v


def fit_nsphere(points, prior=None):
    """
    Fit an n-sphere to a set of points using least squares.

    Parameters
    ------------
    points : (n, d) float
      Points in space
    prior : (d,) float
      Best guess for center of nsphere

    Returns
    ---------
    center : (d,) float
      Location of center
    radius : float
      Mean radius across circle
    error : float
      Peak to peak value of deviation from mean radius
    """
    # make sure points are numpy array
    points = np.asanyarray(points, dtype=np.float64)
    # create ones so we can dot instead of using slower sum
    ones = np.ones(points.shape[1])

    def residuals(center):
        # do the axis sum with a dot
        # this gets called a LOT so worth optimizing
        radii_sq = np.dot((points - center) ** 2, ones)
        # residuals are difference between mean
        # use our sum mean vs .mean() as it is slightly faster
        return radii_sq - (radii_sq.sum() / len(radii_sq))

    if prior is None:
        guess = points.mean(axis=0)
    else:
        guess = np.asanyarray(prior)

    center_result, return_code = leastsq(residuals, guess, xtol=1e-8)

    if return_code not in [1, 2, 3, 4]:
        raise ValueError("Least square fit failed!")

    radii = util.row_norm(points - center_result)
    radius = radii.mean()
    error = np.ptp(radii)
    return center_result, radius, error


def is_nsphere(points):
    """
    Check if a list of points is an nsphere.

    Parameters
    -----------
    points : (n, dimension) float
      Points in space

    Returns
    -----------
    check : bool
      True if input points are on an nsphere
    """
    _center, _radius, error = fit_nsphere(points)
    check = error < tol.merge
    return check



try:
    from . import generic as g
except BaseException:
    import generic as g


def test_fill_holes():
    meshes = [
        "unit_cube.STL",
        "machinist.XAML",
        "round.stl",
        "sphere.ply",
        "teapot.stl",
        "soup.stl",
        "featuretype.STL",
        "angle_block.STL",
        "quadknot.obj",
    ]

    try:
        import cascadio  # noqa
        # if we have cascadio test against a STEP file
        meshes.append("featuretype.STEP")
    except ImportError:
        pass

    for name in meshes:
        mesh = g.get_mesh(name)
        # handle scene lazily
        if hasattr(mesh, "geometry"):
            mesh = next(iter(mesh.geometry.values()))

        if not mesh.is_watertight:
            # output of fill_holes should match watertight status
            returned = mesh.fill_holes()
            assert returned == mesh.is_watertight
            continue

        hashes = [{mesh._data.__hash__(), hash(mesh)}]

        # clip off the first and last few faces
        mask = g.np.ones(len(mesh.faces), dtype=bool)
        mask[[0, -1]] = False
        mesh.update_faces(mask)

        assert not mesh.is_watertight
        assert not mesh.is_volume

        # assert face attributes match faces
        for attrib in mesh.face_attributes.values():
            assert len(attrib) == len(mesh.faces)

        # color some faces
        g.trimesh.repair.broken_faces(mesh, color=[255, 0, 0, 255])

        hashes.append({mesh._data.__hash__(), hash(mesh)})

        assert hashes[0] != hashes[1]

        # put face_normals in cache then corrupt one
        # so we can verify extend_faces preserves cached normals
        _ = mesh.face_normals
        mesh._cache.cache["face_normals"].flags.writeable = True
        mesh._cache.cache["face_normals"][3, :] += 10
        mesh._cache.cache["face_normals"].flags.writeable = False
        cached_marker = mesh.face_normals[3].copy()
        assert (cached_marker >= 9).all()

        # run the fill holes operation should succeed
        assert mesh.fill_holes()
        # should be a superset of the last two
        assert mesh.is_volume
        assert mesh.is_watertight
        assert mesh.is_winding_consistent

        # the corrupted normal should have survived from cache
        assert g.np.allclose(mesh.face_normals[3], cached_marker)

        hashes.append({mesh._data.__hash__(), hash(mesh)})
        assert hashes[1] != hashes[2]

        # assert the filled holes padded attributes
        # this is also the test for `mesh.extend_faces`
        for attrib in mesh.face_attributes.values():
            assert len(attrib) == len(mesh.faces)
        # face colors should have been padded to match
        assert len(mesh.visual.face_colors) == len(mesh.faces)

        # try broken faces on a watertight mesh
        g.trimesh.repair.broken_faces(mesh, color=[255, 255, 0, 255])


def test_fix_normals():
    for mesh in g.get_meshes(5):
        mesh.fix_normals()


def test_winding():
    """
    Reverse some faces and make sure fix_face_winding flips
    them back.
    """

    meshes = [
        g.get_mesh(i)
        for i in [
            "unit_cube.STL",
            "machinist.XAML",
            "round.stl",
            "quadknot.obj",
            "soup.stl",
        ]
    ]

    for i, mesh in enumerate(meshes):
        # turn scenes into multibody meshes
        if g.trimesh.util.is_instance_named(mesh, "Scene"):
            meta = mesh.metadata
            meshes[i] = mesh.dump().sum()
            meshes[i].metadata = meta

    timing = {}
    for mesh in meshes:
        # save the initial state
        is_volume = mesh.is_volume
        winding = mesh.is_winding_consistent

        tic = g.time.time()
        # flip faces to break winding
        mesh.faces[:4] = g.np.fliplr(mesh.faces[:4])

        # run the operation
        mesh.fix_normals()

        # make sure mesh is repaired to former glory
        assert mesh.is_volume == is_volume
        assert mesh.is_winding_consistent == winding

        # save timings
        timing[mesh.source.file_name] = g.time.time() - tic
    # print timings as a warning
    g.log.warning(g.json.dumps(timing, indent=4))


def test_inversion():
    """Make sure fix_inversion switches all reversed faces back"""
    orig_mesh = g.get_mesh("unit_cube.STL")
    orig_verts = orig_mesh.vertices.copy()
    orig_faces = orig_mesh.faces.copy()

    mesh = g.Trimesh(orig_verts, orig_faces[:, ::-1])
    inv_faces = mesh.faces.copy()
    # check not fixed on the way in
    assert not g.np.allclose(inv_faces, orig_faces)

    g.trimesh.repair.fix_inversion(mesh)
    assert not g.np.allclose(mesh.faces, inv_faces)
    assert g.np.allclose(mesh.faces, orig_faces)


def test_multi():
    """
    Try repairing a multibody geometry
    """
    # create a multibody mesh with two cubes
    a = g.get_mesh("unit_cube.STL")
    b = a.copy()
    b.apply_translation([2, 0, 0])
    m = a + b
    # should be a volume: watertight, correct winding
    assert m.is_volume

    # flip one face of A
    a.faces[:1] = g.np.fliplr(a.faces[:1])
    # flip every face of A
    a.invert()
    # flip one face of B
    b.faces[:1] = g.np.fliplr(b.faces[:1])
    m = a + b

    # not a volume
    assert not m.is_volume

    m.fix_normals(multibody=False)

    # shouldn't fix inversion of one cube
    assert not m.is_volume

    # run fix normal with multibody mode
    m.fix_normals()

    # should be volume again
    assert m.is_volume

    # mesh should be volume of two boxes, and positive
    assert g.np.isclose(m.volume, 2.0)


def test_flip():
    # create two spheres
    a = g.trimesh.creation.icosphere()
    b = g.trimesh.creation.icosphere().apply_translation([2, 3, 0])
    # invert the second sphere
    b.faces = g.np.fliplr(b.faces)
    m = a + b
    # make sure normals are in cache
    assert m.face_normals.shape == m.faces.shape
    m.fix_normals(multibody=True)
    assert g.np.isclose(m.volume, a.volume * 2.0)


def test_fan():
    # start by creating an icosphere and removing
    # all faces that include a single vertex to make
    # a nice hole in the mesh
    m = g.trimesh.creation.icosphere()
    clip = m.vertex_faces[0]
    clip = clip[clip >= 0]
    assert len(clip) > 4
    mask = g.np.ones(len(m.faces), dtype=bool)
    mask[clip] = False

    # should have been watertight
    assert m.is_watertight
    assert m.is_winding_consistent
    m.update_faces(mask)
    # now should not be watertight
    assert not m.is_watertight
    assert m.is_winding_consistent

    # create a triangle fan to cover the hole
    stitch = g.trimesh.repair.stitch(m)
    # should be an (n, 3) int
    assert len(stitch.shape) == 2
    assert stitch.shape[1] == 3
    assert stitch.dtype.kind == "i"

    # now check our stitch to see if it handled the hole
    repair = g.trimesh.Trimesh(
        vertices=m.vertices.copy(), faces=g.np.vstack((m.faces, stitch))
    )
    assert repair.is_watertight
    assert repair.is_winding_consistent


def test_fix_normals_mixed_multibox():
    from trimesh.creation import box
    # a test for multibody meshes with some watertight, some not

    # create a multibody mesh with two cubes, one of which is inverted
    a = box()
    b = box().apply_translation([4, 0, 0]).invert()
    c = a + b

    # inverted should sum to zero volume
    assert g.np.isclose(c.volume, 0.0)
    # area isn't signed so should be correct
    assert g.np.isclose(c.area, 12.0)

    # fix_normals should repair the disconnected meshes
    r = c.copy().fix_normals()
    assert g.np.allclose(r.volume, 2.0)

    # create a non-watertight cube
    frag = box().apply_translation([8, 0, 0])
    frag.faces = frag.faces[:-1]
    assert not frag.is_watertight

    # append on the non-watertight cube
    c += frag

    # exactly half of one 1x1 square face is missing
    assert g.np.isclose(c.area, 17.5), c.area

    # run the fix_normals call with:
    # - 1 good cube
    # - 1 inverted cube
    # - 1 cube missing a face
    # this should fix the inverted and ignore the one with holes in it
    c.fix_normals()

    # split without repairing the hole, but culling the non-watertight result
    split = c.split(only_watertight=True, repair=False)
    assert len(split) == 2, len(split)
    assert all(s.is_volume for s in split)
    assert all(g.np.isclose(s.volume, 1.0) for s in split)

    # split WITH repairing the hole and culling anything non-watertight
    split = c.split(only_watertight=True, repair=True)
    assert len(split) == 3, len(split)

    # everything should be a volume
    assert all(s.is_volume for s in split)
    assert all(g.np.isclose(s.volume, 1.0) for s in split)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    test_fill_holes()



import gzip
import json
import os
from io import BytesIO

from ..typed import Dict

# find the current absolute path to this directory
_pwd = os.path.expanduser(os.path.abspath(os.path.dirname(__file__)))
# once resources are loaded cache them
_cache = {}


def get_schema(name: str) -> Dict:
    """
    Load a schema and evaluate the referenced files.

    Parameters
    ------------
    name : str
      Filename of schema.

    Returns
    ----------
    schema
      Loaded and resolved schema.
    """
    from ..resolvers import FilePathResolver
    from ..schemas import resolve

    # get a resolver for our base path
    resolver = FilePathResolver(os.path.join(_pwd, "schema", name))
    # recursively load `$ref` keys
    return resolve(json.loads(resolver.get(name).decode("utf-8")), resolver=resolver)


def get_json(name: str) -> Dict:
    """
    Get a resource from the `trimesh/resources` folder as a decoded string.

    Parameters
    -------------
    name : str
      File path relative to `trimesh/resources/{name}`

    Returns
    -------------
    resource
      File data decoded from JSON.
    """
    raw = get_bytes(name)
    if name.endswith(".gzip"):
        raw = gzip.decompress(raw)
    else:
        raw = raw.decode("utf-8")
    return json.loads(raw)


def get_string(name: str) -> str:
    """
    Get a resource from the `trimesh/resources` folder as a decoded string.

    Parameters
    -------------
    name
      File path relative to `trimesh/resources`

    Returns
    -------------
    resource
      File data as a string.
    """
    return get_bytes(name).decode("utf-8")


def get_bytes(name: str) -> bytes:
    """
    Get a resource from the `trimesh/resources` folder as binary data.

    Parameters
    -------------
    name
      File path relative to `trimesh/resources`

    Returns
    -------------
    resource
      File data as raw bytes.
    """
    cached = _cache.get(name, None)
    if cached is not None:
        return cached

    # get the resource using relative names
    # all templates are using POSIX relative paths
    # so fix them to be platform-specific
    with open(os.path.join(_pwd, *name.split("/")), "rb") as f:
        resource = f.read()

    _cache[name] = resource
    return resource


def get_stream(name: str) -> BytesIO:
    """
    Get a resource from the `trimesh/resources` folder as a binary stream.

    Parameters
    -------------
    name : str
      File path relative to `trimesh/resources`

    Returns
    -------------
    resource
      File data as a binary stream.
    """

    return BytesIO(get_bytes(name))



try:
    from . import generic as g
except BaseException:
    import generic as g

import numpy as np


class DXMLTest(g.unittest.TestCase):
    def test_abaqus_texture(self):
        # an assembly with instancing
        s = g.get_mesh("cube1.3dxml")

        # should be 1 unique meshes
        assert len(s.geometry) == 1

        v = s.geometry["Abaqus_Geometry"].visual
        assert v.kind == "texture"
        assert len(np.unique(v.to_color().vertex_colors, axis=0)) == 4

    def test_abaqus_blocks(self):
        # an assembly with two Faces elements of different color
        s = g.get_mesh("blocks.3dxml")
        assert g.np.isclose(s.volume, 18000, atol=1.0)
        v = s.geometry["Abaqus_Geometry"].visual
        assert v.kind == "face"
        assert len(np.unique(v.face_colors, axis=0)) == 2


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


def test_linear_srgb():
    from trimesh.visual.color import linear_to_srgb, srgb_to_linear, to_rgba

    # deterministically create some colors
    original = g.np.arange(255, dtype=g.np.uint8).reshape((-1, 3))

    # convert them to linear form
    linear = srgb_to_linear(original)

    assert linear.shape == original.shape
    assert g.np.ptp(linear) > 0.0
    assert linear.dtype == g.np.float64

    # convert them back to srgb float colors
    srgb = linear_to_srgb(linear)

    # will convert float -> uint8 and clip off the alpha
    roundtrip = to_rgba(srgb)[:, :3]

    # we should have roundtripped exactly
    assert g.np.allclose(roundtrip, original)


def test_visual():
    mesh = g.get_mesh("featuretype.STL")

    # stl shouldn't have any visual properties defined
    assert not mesh.visual.defined

    for facet in mesh.facets:
        mesh.visual.face_colors[facet] = g.trimesh.visual.random_color()

    assert mesh.visual.defined
    assert not mesh.visual.transparency

    mesh.visual.face_colors[0] = [10, 10, 10, 130]
    assert mesh.visual.transparency


def test_concatenate():
    a = g.get_mesh("ballA.off")
    b = g.get_mesh("ballB.off")

    a.visual.face_colors = [255, 0, 0]
    r = a + b
    assert any(g.np.ptp(r.visual.face_colors, axis=0) > 1)


def test_random_color():
    from trimesh.visual.color import random_color

    c = random_color()
    assert c.shape == (4,)
    assert c.dtype == g.np.uint8

    c = random_color(count=10)
    assert c.shape == (10, 4)
    assert c.dtype == g.np.uint8


def test_hsv_rgba():
    # the non-vectorized stdlib HSV -> RGB function
    from colorsys import hsv_to_rgb

    # our HSV -> RGBA function
    from trimesh.visual.color import hsv_to_rgba

    # create some random HSV values in the 0.0 - 1.0 range
    hsv = g.random((100, 3))

    # run our conversion
    ours = hsv_to_rgba(hsv, dtype=g.np.float64)

    # check the result from the standard library
    truth = g.np.array([hsv_to_rgb(*v) for v in hsv])

    # they should match
    assert g.np.allclose(ours[:, :3], truth, atol=0.0001), ours[:, :3] - truth


def test_to_rgba_float():
    a = g.np.zeros((100, 4), dtype=g.np.float64)
    c = g.trimesh.visual.to_rgba(a)
    assert g.np.allclose(c, 0)
    assert c.dtype == g.np.uint8
    # should be clipped to full-opaque
    a[:, 3] = 10.0
    c = g.trimesh.visual.to_rgba(a)
    assert g.np.allclose(c[:, :3], 0.0)
    assert g.np.allclose(c[:, 3], 255)
    assert c.dtype == g.np.uint8

    a = g.np.ones((100, 4), dtype=g.np.float64)
    c = g.trimesh.visual.to_rgba(a)
    assert g.np.allclose(c, 255)
    assert c.dtype == g.np.uint8


def test_concatenate_empty_mesh():
    box = g.get_mesh("box.STL")

    mesh_fcolor = box.copy()
    mesh_fcolor.visual.face_colors = [255, 0, 0]

    mesh_vcolor = box.copy()
    mesh_vcolor.visual.vertex_colors = [0, 0, 255]

    mesh_empty = g.trimesh.Trimesh()

    r_left_fcolor = mesh_fcolor + mesh_empty
    r_right_fcolor = mesh_empty + mesh_fcolor
    r_left_vcolor = mesh_vcolor + mesh_empty
    r_right_vcolor = mesh_empty + mesh_vcolor
    r_empty = mesh_empty + mesh_empty

    for visual_face in [r_left_fcolor.visual, r_right_fcolor.visual]:
        assert (visual_face.face_colors == mesh_fcolor.visual.face_colors).all()
        assert visual_face.kind == "face"

    for visual_vert in [r_left_vcolor.visual, r_right_vcolor.visual]:
        assert (visual_vert.vertex_colors == mesh_vcolor.visual.vertex_colors).all()
        assert visual_vert.kind == "vertex"

    assert len(r_empty.visual.face_colors) == 0
    assert r_empty.visual.kind is None


def test_data_model():
    """
    Test the probably too- magical color caching and storage
    system.
    """
    m = g.get_mesh("featuretype.STL")
    test_color = [255, 0, 0, 255]
    test_color_2 = [0, 255, 0, 255]
    test_color_transparent = [25, 33, 0, 146]

    # there should be nothing in the cache or DataStore when
    # starting
    assert len(m.visual._cache) == 0
    assert len(m.visual._data) == 0
    # no visuals have been defined so this should be None
    assert m.visual.kind is None
    assert not m.visual.defined

    # this should cause colors to be generated into cache
    initial_id = id(m.visual.face_colors)
    assert m.visual.face_colors.shape[0] == len(m.faces)
    assert id(m.visual.face_colors) == initial_id
    # the values should be in the cache and not in data
    assert len(m.visual._cache) > 0
    assert len(m.visual._data) == 0
    assert not m.visual.defined
    assert not m.visual.transparency

    # this should move the color from cache to data
    m.visual.face_colors[0] = test_color
    # the operation should have moved the colors into data but
    # the object ID should be the same as on creation
    # assert id(m.visual.face_colors) == initial_id
    # the color assignment inside the array should have worked
    assert (m.visual.face_colors[0] == test_color).all()
    # the rest of the colors should be unchanged
    assert (m.visual.face_colors[1] != test_color).any()
    assert len(m.visual._data) >= 1
    assert m.visual.kind == "face"
    assert m.visual.defined
    assert not m.visual.transparency

    # set all face colors to test color
    m.visual.face_colors = test_color
    assert (m.visual.face_colors == test_color).all()
    # assert len(m.visual._cache) == 0
    # should be just material and face information
    assert len(m.visual._data.data) >= 1
    assert m.visual.kind == "face"
    assert bool((m.visual.vertex_colors == test_color).all())
    assert m.visual.defined
    assert not m.visual.transparency

    # this should move the color from cache to data
    m.visual.vertex_colors[0] = test_color_2
    assert (m.visual.vertex_colors[0] == test_color_2).all()
    assert (m.visual.vertex_colors[1] != test_color_2).any()
    assert m.visual.kind == "vertex"
    assert m.visual.defined
    assert not m.visual.transparency

    m.visual.vertex_colors[1] = test_color_transparent
    assert m.visual.transparency

    test = (g.random((len(m.faces), 4)) * 255).astype(g.np.uint8)
    m.visual.face_colors = test
    assert bool((m.visual.face_colors == test).all())
    assert m.visual.kind == "face"

    test = (g.random((len(m.vertices), 4)) * 255).astype(g.np.uint8)
    m.visual.vertex_colors = test
    assert bool((m.visual.vertex_colors == test).all())
    assert m.visual.kind == "vertex"

    test = (g.random(4) * 255).astype(g.np.uint8)
    m.visual.face_colors = test
    assert bool((m.visual.vertex_colors == test).all())
    assert m.visual.kind == "face"
    m.visual.vertex_colors[0] = [0, 0, 0, 0]
    assert m.visual.kind == "vertex"

    test = (g.random(4) * 255).astype(g.np.uint8)
    m.visual.vertex_colors = test
    assert bool((m.visual.face_colors == test).all())
    assert m.visual.kind == "vertex"
    m.visual.face_colors[:2] = (g.random((2, 4)) * 255).astype(g.np.uint8)
    assert m.visual.kind == "face"


def test_smooth():
    """
    Make sure cached smooth model is dumped if colors are changed
    """
    m = g.get_mesh("featuretype.STL")

    # will put smoothed mesh into visuals cache
    s = m.smooth_shaded
    # every color should be default color
    assert g.np.ptp(s.visual.face_colors, axis=0).max() == 0

    hash_pre = hash(m.visual)
    # set one face to a different color
    m.visual.face_colors[0] = [255, 0, 0, 255]

    # cache should be dumped yo
    s1 = m.smooth_shaded

    assert hash(m.visual) != hash_pre

    assert g.np.ptp(m.visual.face_colors, axis=0).max() != 0
    assert g.np.ptp(s1.visual.face_colors, axis=0).max() != 0

    # do the same check on vertex color
    m = g.get_mesh("featuretype.STL")
    s = m.smooth_shaded
    # every color should be default color
    assert g.np.ptp(s.visual.vertex_colors, axis=0).max() == 0
    m.visual.vertex_colors[g.np.arange(10)] = [255, 0, 0, 255]
    s1 = m.smooth_shaded
    assert g.np.ptp(s1.visual.face_colors, axis=0).max() != 0


def test_vertex():
    m = g.get_mesh("torus.STL")

    m.visual.vertex_colors = [100, 100, 100, 255]

    assert len(m.visual.vertex_colors) == len(m.vertices)


def test_conversion():
    m = g.get_mesh("machinist.XAML")
    assert m.visual.kind == "face"

    # unmerge vertices so we don't get average colors
    m.unmerge_vertices()

    # store initial face colors
    initial = g.deepcopy(m.visual.face_colors.copy())

    # assign averaged vertex colors as default
    m.visual.vertex_colors = m.visual.vertex_colors
    assert m.visual.kind == "vertex"

    m.visual._cache.clear()
    assert g.np.allclose(initial, m.visual.face_colors)


def test_interpolate():
    """
    Check our color interpolation
    """
    values = g.np.array([-1.0, 0.0, 1.0, 2.0])
    # should clamp
    colors = g.trimesh.visual.linear_color_map(values)

    assert g.np.allclose(colors[0], [255, 0, 0, 255])
    assert g.np.allclose(colors[1], [255, 0, 0, 255])
    assert g.np.allclose(colors[2], [0, 255, 0, 255]), colors[2]
    assert g.np.allclose(colors[3], [0, 255, 0, 255])

    # make sure it's interpolating
    colors = g.trimesh.visual.linear_color_map([0.0, 0.5, 1.0])
    assert g.np.allclose(colors[0], [255, 0, 0, 255])
    assert g.np.allclose(colors[1], [128, 128, 0, 255]), colors[1]
    assert g.np.allclose(colors[2], [0, 255, 0, 255])

    # should scale to range
    colors = g.trimesh.visual.interpolate(
        values, color_map=g.trimesh.visual.linear_color_map
    )
    assert g.np.allclose(colors[0], [255, 0, 0, 255])
    # scaled to range not clamped
    assert not g.np.allclose(colors[1], [255, 0, 0, 255])
    assert not g.np.allclose(colors[2], [0, 255, 0, 255])
    # end of range
    assert g.np.allclose(colors[3], [0, 255, 0, 255])

    # try interpolating with matplotlib color maps

    colors = g.trimesh.visual.interpolate(values, "viridis")

    # check shape and type for matplotlib cmaps
    assert colors.shape == (len(values), 4)
    assert colors.dtype == g.np.uint8
    # every color should differ
    assert (colors[:-1] != colors[1:]).any(axis=1).all()

    # make sure it handles zero range
    # use the base linear_color_map which assigns 0 -> red
    colors = g.trimesh.visual.interpolate(
        g.np.zeros(100), color_map=g.trimesh.visual.color.linear_color_map
    )
    assert g.np.allclose(colors, [255, 0, 0, 255])

    # create a subdivided box and interpolate the radius
    box = g.trimesh.creation.box().subdivide(iterations=5)
    radii = g.np.linalg.norm(box.vertices, axis=1)
    # get the order for comparison
    order = radii.argsort()

    # run the interpolation with a pure red-green
    box.visual.vertex_colors = g.trimesh.visual.color.interpolate(
        radii, color_map=g.trimesh.visual.color.linear_color_map
    )
    # box.show(smooth=False)

    # extract colors in order
    color = box.visual.vertex_colors[order]

    # green should be going from 0-255
    # change the dtype so `diff` doesn't wrap
    green = color[:, 1].astype(g.np.int64)

    # first value is zero
    assert green[0] == 0
    # last value is fully set
    assert green[-1] == 255

    # should be completely monotonic
    diff = green[1:] - green[:-1]
    assert (diff >= 0).all(), diff.min()

    # should be pretty smooth
    assert diff.max() < 15, diff.max()

    # now make a box with viridis vertex colors that you can
    # add a `box.show()` to if you want to see if it actually does something
    box.visual.vertex_colors = g.trimesh.visual.interpolate(radii, "viridis")
    # make sure colors aren't all the same
    assert g.np.ptp(box.visual.vertex_colors, axis=0).max() > 0

    # now see if we match matplotlib if it's installed
    try:
        from matplotlib.pyplot import get_cmap
    except ImportError:
        return

    # make sure a `callable` works
    check = g.trimesh.visual.interpolate(radii, color_map=get_cmap("viridis"))

    # `get_cmap` doesn't interpolate linearly but otherwise "their viridis"
    # and "our viridis" should be decently close
    assert g.np.allclose(box.visual.vertex_colors, check, atol=3)

    # box.show()


def test_uv_to_color():
    try:
        import PIL.Image
    except ImportError:
        return

    # n_vertices = 100
    uv = g.np.array([[0.25, 0.2], [0.4, 0.5]], dtype=float)
    texture = g.np.arange(96, dtype=g.np.uint8).reshape(8, 4, 3)
    colors = g.trimesh.visual.uv_to_color(uv, PIL.Image.fromarray(texture))
    colors_expected = [[75, 76, 77, 255], [51, 52, 53, 255]]

    g.np.testing.assert_allclose(colors, colors_expected, rtol=0, atol=0)


def test_uv_to_interpolated_color():
    try:
        import PIL.Image
    except ImportError:
        return

    uv = g.np.array([[0.25, 0.2], [0.4, 0.5]], dtype=float)

    tmp = g.np.arange(32)
    x, y = g.np.meshgrid(tmp, tmp)
    z = x + 128

    texture = g.np.stack([x, y, z], axis=-1).astype(g.np.uint8)

    img = PIL.Image.fromarray(texture)
    colors = g.trimesh.visual.uv_to_interpolated_color(uv, img)

    # exact interpolated values before being converted to uint8
    colors_expected = [[7.75, 24.8, 128 + 7.75, 255], [12.4, 15.5, 128 + 12.4, 255]]

    assert g.np.allclose(colors, colors_expected, rtol=0, atol=1)


def test_iterset():
    m = g.trimesh.creation.box()
    color = [100, 0, 0, 200]

    # facets should include every face
    assert set(g.np.hstack(m.facets)) == set(range(len(m.faces)))
    assert len(m.facets) * 2 == len(m.faces)

    for f in m.facets:
        m.visual.face_colors[f] = color

    assert g.np.allclose(m.visual.face_colors, color)


def test_copy():
    s = g.trimesh.creation.uv_sphere().scene()
    s.geometry["geometry_0"].visual.face_colors[:, :3] = [0, 255, 0]

    a = s.geometry["geometry_0"]
    assert id(a) == id(s.geometry["geometry_0"])

    b = s.geometry["geometry_0"].copy()
    assert id(a) != id(b)

    assert g.np.allclose(a.visual.face_colors, b.visual.face_colors)


if __name__ == "__main__":
    # test_to_rgba_float()
    # test_interpolate()
    test_linear_srgb()
